//--------------------------------------------------------------------------
// w3dCamera.js - Handhabung der Camera und Viewpoints           2016-02-21

var curCamera;
var curCamName;

var camXposDst, camYposDst, camZposDst;
var camXrotDst, camYrotDst, camZrotDst;
var camXpos, camYpos, camZpos;
var camXrot, camYrot, camZrot;

var yAugen = 1.5;               // Default-Augenhoehe
var dSide = 0.0;
var sDirec = false;

//--------------------------------------------------------------------------

function initCamera() {

    curCamera = 0;
    curCamName = "FPV";
    camXposDst = camXpos = posX;    // Position ist schon bekannt
    camYposDst = camYpos = posY;
    camZposDst = camZpos = posZ;
    camXrotDst = camXrot = rotX*Math.PI/180;
    camYrotDst = camYrot = rotY*Math.PI/180;
    camZrotDst = camZrot = rotZ*Math.PI/180;
}

//--------------------------------------------------------------------------
// Zyklische Nachfuehrung der Szenenkamera

function renderCamera() {

    if (viewMode == 0) {            // Normalsicht
        camera.position.set(camXpos+dSide*dzMove, camYpos+yAugen, camZpos-dSide*dxMove);
        camera.rotation.set(camXrot, -camYrot, camZrot);
    } else if (viewMode == 1) {     // 3rd Person View
        camera.position.set(camXpos+5*dxMove, camYpos+yAugen + 1.3, camZpos+5*dzMove);
        camera.lookAt(new THREE.Vector3(posX, posY+yAugen, posZ));
    } else if (viewMode == 2) {     // BirdsEye View
        camera.position.set(camXpos+80*dxMove, camYpos + 80, camZpos+80*dzMove);
        camera.lookAt(new THREE.Vector3(posX, posY+yAugen, posZ));
    } else {                        // Satellite View
        camera.position.set(camXpos-20, 220, camZpos-180);
        camera.lookAt(new THREE.Vector3(posX, posY+yAugen, posZ));
    }
}

//--------------------------------------------------------------------------
// Position und Ausrichtung der Kamera nachfuehren; bei Ortsveraenderungen
// mehr oder weniger schnelles Hinuebergleiten.

function cycleCamera(msPerFrame) {

    var cFkt = 0.0025;
    if (curCamera == 0) {           // First Person View
        cFkt = 0.01;                // schneller nachfuehren
        camXposDst = posX;
        camYposDst = posY;
        camZposDst = posZ;
        camXrotDst = rotX*Math.PI/180;
        camYrotDst = rotY*Math.PI/180;
        camZrotDst = rotZ*Math.PI/180;
    }

    if (camYrotDst > 2*Math.PI)
        camYrotDst -= 2*Math.PI;
    else if (camYrotDst < 0)
        camYrotDst += 2*Math.PI;

    var dwi = camYrotDst - camYrot;
    if (dwi > Math.PI)
        dwi -= 2*Math.PI;
    else if (dwi < -Math.PI)
        dwi += 2*Math.PI;

    var dFkt = cFkt * msPerFrame;
    camXpos += (camXposDst - camXpos) * dFkt;
    camYpos += (camYposDst - camYpos) * dFkt;
    camZpos += (camZposDst - camZpos) * dFkt;
    camXrot += (camXrotDst - camXrot) * dFkt;
    camYrot += dwi * dFkt;
//    camYrot = camYrotDst;
    camZrot += (camZrotDst - camZrot) * dFkt;

    skyBox.position.set(camXpos, camYpos, camZpos);

    // EXPERIMENT: leichtes Hin- und Herbewegen - funktioniert, nervt aber evt.
    if ((mUseSensors == 0) && (curCamera == 0)) {
        var dst = msPerFrame * 0.00004;
        if (sDirec) {
            if (dSide > -0.012)
                dSide -= dst;
            else
                sDirec = false;
        } else {
            if (dSide < 0.012)
                dSide += dst;
            else
                sDirec = true;
        }
    }
}

//--------------------------------------------------------------------------

function changeCamera(cam, num) {

    curCamera = num;
    curCamName = cam[0];
    camXposDst = cam[1];
    camYposDst = cam[2];
    camZposDst = cam[3];
    camXrotDst = cam[4];
    camYrotDst = cam[5];
    camZrotDst = cam[6];

    if (curCamera == 0)
        setTileOffsets();
}

