//--------------------------------------------------------------------------
// w3dTiles.js - Laden und Verwalten der Tile-Scripts           2016-02-21

// Je geladener Kachel eine Position in den folgenden Datenlisten:

var MAX_TILES = 50;                         // bis zu 9 Kacheln im Speicher + Reserve

var qVisible = new Array(MAX_TILES);        // Sichtbarkeit der Tiles
var qTiles = new Array(MAX_TILES);          // Tile-Objekte
var qIds = new Array(MAX_TILES);            // Tile-IDs
var qNames = new Array(MAX_TILES);          // Tile-Namen
var qXofs = new Array(MAX_TILES);           // x-Offset je Tile
var qZofs = new Array(MAX_TILES);           // z-Offset je Tile
var qObjects = new Array(MAX_TILES);        // Szenen-Summenobjekte
var qBBoxes = new Array(MAX_TILES);         // BoundingBoxes
var qBBoxFrames = new Array(MAX_TILES);     // Wireframes dazu

var curTid;                                 // aktuelle Tile-ID
var curIdx;                                 // Index akt. Tile in der Liste

for (var i = 0; i < MAX_TILES; i++) {
    qNames[i] = "";
    qVisible[i] = false;
}


// Je VSC-Script eine Position in den folgenden Datenlisten:

var MAX_VSC = 50;                           // max Anzahl aktive VSC-Scripts

var vFilename = new Array(MAX_VSC);         // Filename der VSC-Scripts
var vState = new Array(MAX_VSC);            // Status: 0-passiv, 1-angef., 2-registriert
var vScript = new Array(MAX_VSC);           // Inhalt des VSC-Scripts
var vTileId = new Array(MAX_VSC);           // zugehoerige Tile-ID je VSC-Script
var vObjects = new Array(MAX_VSC);          // Three.js-Summenobjekt je VSC-Script 
var vVersions = new Array(MAX_VSC);         // Versionskennung
var vVisible = new Array(MAX_VSC);          // Sichtbarkeit der VSC-Objekte
var vBBoxes = new Array(MAX_VSC);           // Bounding-Boxes
var vBBoxFrames = new Array(MAX_VSC);       // Wireframes dazu
var vSymbols = new Array(MAX_VSC);          // Symboltabelle
var vDynamics = new Array(MAX_VSC);         // Dynamik-Tabelle
var vSensors = new Array(MAX_VSC);          // Sensorpunkte-Tabelle
var vCameras = new Array(MAX_VSC);          // Viewpoint-Tabelle
var vTracks = new Array(MAX_VSC);           // Strassen/Gleise
var vTrackObjs = new Array(MAX_VSC);        // Three.js-Objekt dazu
var vBuildTime = new Array(MAX_VSC);        // Kompilierdauer in ms

for (var i = 0; i < MAX_VSC; i++) {
    vFilename[i] = "";
    vScript[i] = "";
    vState[i] = 0;
    vTileId[i] = 0;
    vVisible[i] = false;
}

var vnQueue = new Array(MAX_VSC);
var vnRdIdx = 0;
var vnWrIdx = 0;
var bVscBusy = false;
var BusyTimeout = Date.now();
var cntRetry = 0;


var jsSumText = "";
var jsElement = undefined;
var bSceneValid = false;
var loadState = 0;
var GlobalList = undefined;                 // Tabelle globaler Symbole


// Fuer die aktuell sichtbaren Tiles gueltige Daten; Koordinaten absolut

var tCamList = undefined;                    // aktuell verfuegbare Kameras
var tSymbList = undefined;                   // aktuell verfuegbare Symbole
var tSensList = undefined;                   // aktuell gueltige Sensoren
var tDynaList = undefined;                   // aktuell gueltige Dynamics



function setTileOffsets() {
    ofsX = Math.floor((posX+LEN_QUAD/2)/LEN_QUAD);
    ofsZ = Math.floor((posZ+LEN_QUAD/2)/LEN_QUAD);
    var qid = (50+ofsX)*100 + (50+ofsZ);
    if (qid == curTid)
        return false;                   // keine Aenderung
    curTid = qid;
    curIdx = idToNum(qid);
    return true;                        // andere Kachel betreten
}

//-----------------------------------------------------------------------------
// Zu einer Kachel-ID den Tabellenplatz ermitteln.

var idToNum = function(qid) {
    if (qid == 0)
        return -1;                      // Tile-ID undefiniert
    for (var i = 0; i < MAX_TILES; i++) {
        if (qIds[i] == qid)
            return i;
    }
    return -1;                          // Tile ist nicht enthalten
};

//-----------------------------------------------------------------------------
// Alle Tiles, deren JS-Code geladen wurde, als Objekte erzeugen

function enterTilesToHTML(jsText) {

    var heads = document.getElementsByTagName("head");
    if (jsElement !== undefined) {
        heads[0].removeChild(jsElement);
        delete jsElement;
    }
    jsElement = document.createElement("script");
    jsElement.type = "text/javascript";
    heads[0].appendChild(jsElement);
    jsElement.textContent = jsText;     // besser als innerHTML
    jsSumText = jsText;

    var nNew = 0;
    var text = jsText;
    while (text.length > 0) {
        var n1 = text.indexOf("//@@Tile ");
        if (n1 < 0)
            break;
        n1 = text.indexOf("var w", n1);
        if (n1 < 0)
            break;
        n1 += 4;
        var qName = text.slice(n1, n1+5);
        n1 = text.indexOf("//@@END");
        if (n1 < 0)
            break;
        text = text.slice(n1+10, text.length);

        var bExist = false;
        for (var i = 0; i < MAX_TILES; i++) {
            if (qName == qNames[i]) {
                bExist = true;
                break;
            }
        }
        if (bExist)                     // JS-Code bereits im Speicher
            continue;

        var tn;
        for (tn = 0; tn < MAX_TILES; tn++) {
            if (qNames[tn] == "")       // unbelegter Platz
                break;
        }
        if (tn >= MAX_TILES) {
            console.error("*** MAX_TILES ueberschritten");
            break;
        }

        qNames[tn] = qName;
        var qid = parseInt(qName.slice(1));
        qIds[tn] = qid;
        qXofs[tn] = (Math.floor(qid/100)-50) * LEN_QUAD;
        qZofs[tn] = (qid%100-50) * LEN_QUAD;

        var cmd = 'new ' + qName + '();';
        var tile = eval(cmd);           // Kachel-Objekt erzeugen
        if (tile === undefined) { 
            console.error ("Tile undefined: "+qName);
            break;
        }
        qTiles[tn] = tile;
        qVisible[tn] = false;            // zunaechst nicht sichtbar
if (LOG>1) console.log("JS Create: "+tn+", "+qName+" - "+tile.qGetSceneName);

        if (qid == curTid)              // aktuelle Kachel?
            curIdx = tn;
        nNew++;
    }
    bSceneValid = true;
if (LOG>2) console.log("--> Created "+nNew+" Tiles");

    loadCurrentScene(true);         // Liste der VSC-Filenamen besorgen
    if (GlobalList === undefined)
        ReadGlobalSymbols();        // Liste globaler Kameras und Symbole
}

//-----------------------------------------------------------------------------
// Ueberfluessig gewordenen JS-Code aus dem Speicher entfernen.

function removeTileData() {

    var nDel = 0;
    for (var tn = 0; tn < MAX_TILES; tn++) {
        if (qNames[tn] == "")           // unbelegter Platz
            continue;
        if (qTiles[tn] !== undefined)   // wird noch gebraucht
            continue;

        // zu loeschen: JS-Code aus der HTML-Struktur entfernen
        var seekTxt = "var " + qNames[tn] + " = function()";
        var n1 = jsSumText.indexOf(seekTxt);
        if (n1 >= 0) {
            var n2 = jsSumText.indexOf("//@@END@@");
            if (n2 < n1)
                console.error("*** JS-Endekennung nicht zu finden: "+qNames[tn]);
            else
                jsSumText = jsSumText.slice(0, n1) + jsSumText.slice(n2+10);
        }
        qNames[tn] = "";
        nDel++;
    }
    jsElement.innerHTML = jsSumText;    //@@ reicht das ??
if ((LOG>1) && (nDel > 0)) console.log("JS Deleted "+nDel+" Tiles");
}

//-----------------------------------------------------------------------------
// Nachladen von JS-Scripts fuer 9 benachbarte Kacheln anfordern.
// Die Kachel der aktuellen Position ist die mittlere.

function loadCurrentScene(bVscList) {

//    bSceneValid = false;
    setTileOffsets();

    var xmlhttp = new XMLHttpRequest();
    xmlhttp.onreadystatechange = function() {
        if ((xmlhttp.readyState == 4) && (xmlhttp.status == 200)) {
            var rxText = xmlhttp.responseText;
            if (rxText.slice(0, 5) == "quad|") {
                var jsText = rxText.slice(5, rxText.length);  
                enterTilesToHTML(jsText); 
                removeTileData();               
            } else if (rxText.slice(0, 8) == "vsclist|") {
                registerVscFilenames(rxText.slice(8));
            } else {
                console.error("xquad.php: "+rxText);
            }
        }
    };

    inhalt = "../php/xquad.php?quad=" + curTid;
    if (bVscList !== undefined)
        inhalt = inhalt + "&vsclist=1";
    xmlhttp.open("GET", inhalt, true);
    xmlhttp.send();
}

//-----------------------------------------------------------------------------
// Eine bestimmte Kachel sichtbar machen, ggf. neu einrichten

function doShowScene(qid) {

    var tn = idToNum(qid);
    if (tn < 0) {                       // Kachel nicht geladen?
        loadCurrentScene();             // Kontext neu besorgen
        return true;
    }

    if (bVscBusy && (Date.now() > BusyTimeout)) {
        bVscBusy = false;
        if (cntRetry++ < 3) {           // Wiederholung versuchen
if (LOG) console.log("--> Timeout nach VSC-Request, Retry "+cntRetry);
            // der Queue-Lesezeiger steht noch auf dem letzten Eintrag
            vscSendReq(vFilename[vnRdIdx], vTileId[vnRdIdx]);
            return false;
        }
if (LOG) console.log("*** Timeout nach erfolglosen Retries");
    }

    if ((vnRdIdx != vnWrIdx) && !bVscBusy) {
        var vn = vnQueue[vnRdIdx];
        if (++vnRdIdx >= MAX_VSC)
            vnRdIdx = 0;
        vscSendReq(vFilename[vn], vTileId[vn]);
        vState[vn] = 2;
    }
    
    var bNew = false;
    if (!qVisible[tn]) {                // Kachel-JS noch unsichtbar?
        if (qObjects[tn] === undefined) {// Tile-Objekt noch zu bauen
            if (qTiles[tn] === undefined) {
                console.error("*** qTiles[w"+qid+"] undefined");
                return false;
            }
            var obj = qTiles[tn].qBuildScene(qid);
if (LOG>1) console.log("JS Build: w"+qid+":"+qTiles[tn].qGetSceneName);
            if (obj === undefined) {
                console.error("buildScene() failed: w"+qid);
                return false;
            }
            obj.position.set(qXofs[tn], 0, qZofs[tn]);
            qObjects[tn] = obj;
            scene.add(obj);
            qVisible[tn] = true;
            bNew = true;                // neu gebaut: danach warten
        }
    }
    for (var vn = 0; vn < MAX_VSC; vn++) {
        if (!vVisible[vn] && (vFilename[vn] != "") && (vTileId[vn] == qid))
            doShowVsc(vn);
    }
    return bNew;
}

function doShowVsc(vn) {
    if (vObjects[vn] !== undefined) {
        scene.add(vObjects[vn]);
        if (bShowBoundingBoxes)
            scene.add(vBBoxFrames[vn]);
        if (vTrackObjs[vn] !== undefined)
            scene.add(vTrackObjs[vn]);
        vVisible[vn] = true;
if (LOG>1) console.log("vsc Visible: "+vFilename[vn]);
    }
}

function doHideVsc(vn) {
    if (vObjects[vn] === undefined)
        return;
    if (vVisible[vn]) {
        vVisible[vn] = false;
        scene.remove(vObjects[vn]);
        if (bShowBoundingBoxes)
            scene.remove(vBBoxFrames[vn]);
        if (vTrackObjs[vn] !== undefined)
            scene.remove(vTrackObjs[vn]);
    }
}

function doTerminateVsc(vn) {
    if (vObjects[vn] === undefined)
        return;
    doHideVsc(vn);
    delete vObjects[vn];
    vObjects[vn] = undefined;
    delete vBBoxFrames[vn];
    vBBoxFrames[vn] = undefined;
    if (vTrackObjs[vn] !== undefined) {
        delete vTrackObjs[vn];
        vTrackObjs[vn] = undefined;
    }
}

//-----------------------------------------------------------------------------
// Eine bestimmte Kachel verbergen aber im Speicher belassen

function doHideScene(qid) {

    var tn = idToNum(qid);
    if (qVisible[tn]) {                 // noch sichtbar?
        qVisible[tn] = false;
        if (qObjects[tn] !== undefined)
            scene.remove(qObjects[tn]);
if (LOG>1) console.log("JS Hide: w"+qid);
        for (var vn = 0; vn < MAX_VSC; vn++) {
            if (vTileId[vn] == qid) {
                doHideVsc(vn);
if (LOG>1) console.log("vsc Hide: w"+qid+":"+vFilename[vn]);
            }
        }
    }
}

//-----------------------------------------------------------------------------
// Eine bestimmte Kachel terminieren, darf danach geloescht werden.

function doTerminateScene(qid) {

    doHideScene(qid);                   // bei Spruengen evt. noetig
    var tn = idToNum(qid);
    if (qTiles[tn] === undefined)
        return;

if (LOG>1) console.log("JS Terminate: w"+qid);
    qTiles[tn].qTerminateScene();
    delete qTiles[tn];
    qTiles[tn] = undefined;
    delete qObjects[tn];
    qObjects[tn] = undefined;
    for (var i = 0; i < MAX_VSC; i++) {
        if (vTileId[i] == qid) {
            doTerminateVsc(i);
            vTileId[i] = 0;
        }
    }
}

//-----------------------------------------------------------------------------
// Einen VSC-Filenamen in die Liste der aktuellen Scripts aufnehmen

function vscRequest(name, qid) {

    if (_DEBUG) {
       if (qid != curTid)
            return;
    }

if (LOG) console.log("vscRequest: w"+qid+":"+name);
    var vn;
    for (vn = 0; vn < MAX_VSC; vn++) {
        if (vFilename[vn] == name)      // VSC ist bereits erfasst
            break;
    }
    if (vn >= MAX_VSC) {                // neues VSC-File: Platz suchen
        for (vn = 0; vn < MAX_VSC; vn++) {
            if (vFilename[vn] != "")
                continue;
            vFilename[vn] = name;
            vTileId[vn] = qid;
            break;
        }
    }
    if (vn >= MAX_VSC) {
        alert("*** vscReqest: MAX_VSC ueberschritten");
        return;
    }
    vState[vn] = 1;                     // Laden des Codes anfordern
    vnQueue[vnWrIdx] = vn;
    if (++vnWrIdx >= MAX_VSC)
        vnWrIdx = 0;
}

//-----------------------------------------------------------------------------
// Antwort auf die Anforderung der aktuellen VSC-Scripts auswerten.

function registerVscFilenames(namText) {

if (LOG>2) console.log("registerVscFilenames: " + namText);
    for (var pos = 0; pos < namText.length; ) {
        pos = namText.indexOf(":", pos);
        if (pos < 5)
            break;
        var qid = parseInt(namText.slice(pos-4, pos));
        pos++;
        var pos1 = namText.indexOf(";", pos);
        if (pos1 < 0)
            break;
        var fName = namText.slice(pos, pos1);
        if (vscRequest(fName, qid) < 0) {
            console.error("*** registerVscFilenames(): MAX_VSC ueberschritten");
            break;
        }
    }
}

//-----------------------------------------------------------------------------------
// VSC-Script anfordern, Inhalt kommt dann asynchron ueber recvSceneData().
// vscName angeben, wenn gezielt ein bestimmtes File zu laden ist, ansonsten der
// Server-Seite ueberlassen, was zu aktualisieren ist.

function vscSendReq(vscName, qid) {

    xmlhttp = new XMLHttpRequest();
    xmlhttp.onreadystatechange = function() {
        if ((xmlhttp.readyState == 4) && (xmlhttp.status == 200)) {
            var rxText = xmlhttp.responseText;
            if ((rxText.length > 6) && (rxText.indexOf("w") == 0)) {
                vscRecvData(rxText);
            } else if (rxText.length > 2)
                console.error("xreadvsc.php: "+rxText);
        }
    };
    var inhalt = "../php/xreadvsc.php?qname=w" + qid + "&uid=" + MyUserId;
    if (vscName == "")                  // nach VSC-Aenderungen pollen?
        inhalt = inhalt + "&poll=1";
    else {                              // ein bestimmtes Script laden
        inhalt = inhalt + "&vsc=" + vscName;
        bVscBusy = true;
        BusyTimeout = Date.now() + 8000;
if (LOG) console.log("vscSendReq:  w"+qid+":"+vscName);
    }
    xmlhttp.open("GET", inhalt, true);  // asynchron
    xmlhttp.send();
}

function vscRecvData(rxText) {
    var qname = rxText.slice(0, 5);
    var pos = rxText.indexOf("|", 6);
    if (pos <= 0) {
        alert("*** vscRecvData: Formatfehler");
        return;
    }
    var fName = rxText.slice(6, pos);
    pos++;
if (LOG) console.log("vscRecvData: "+qname+":"+fName);
    var qid = parseInt(qname.slice(1));
    for (var vn = 0; vn < MAX_VSC; vn++) {
        if ((vTileId[vn] == qid) && (vFilename[vn] == fName)) {
            vScript[vn] = rxText.slice(pos);
            registerVscData(qid, vn);
            break;
        }
    }
}

//-----------------------------------------------------------------------------
// Ergebnisse einer VSC-Bearbeitung auswerten und Datenlisten aufbauen.

function registerVscData(qid, vn) {

    if ((vState[vn] != 2) && (vState[vn] != 3)) {
if (LOG) console.log("*** Unerwarteter Status: "+vState[vn]);
    }
    var tim = Date.now();               // Messung der Kompilierzeit
    var vscData = buildVscScene(vFilename[vn], vScript[vn]);
    vBuildTime[vn] = Date.now() - tim;

if (LOG>1) {
  var str = "";
  if (vState[vn] == 3)
    str = " RELOAD";
  console.log("registerVscData: w"+qid+":"+vscData[0]+" ("+vBuildTime[vn]+"ms)"+str);
}
    var tn = idToNum(qid);              // Nummer in Tile-Liste
    // vorherigen Stand ggf. entfernen und loeschen
    doTerminateVsc(vn);

    // neue Daten auf die diversen Listen verteilen
    vVersions[vn] = vscData[1];
    var vscObj = vscData[2];
    vscObj.position.set(qXofs[tn], 0, qZofs[tn]);
    vObjects[vn] = vscObj;
    vVisible[vn] = false;

    var bbox = vscData[3];
    handleBBoxOffsets(tn, bbox);
    vBBoxes[vn] = bbox;

    var bbFrames = vscData[4];
    bbFrames.position.set(qXofs[tn], 0, qZofs[tn]);
    vBBoxFrames[vn] = bbFrames;

    vSymbols[vn] = vscData[5];
    vDynamics[vn] = vscData[6];
    vSensors[vn] = vscData[7];
    vCameras[vn] = vscData[8];
    vTracks[vn] = vscData[9];

    if (vTracks[vn].length > 0) {
        var trObjs = buildTrackObject(vTracks[vn]);
        trObjs.position.set(qXofs[tn], 0, qZofs[tn]);
        vTrackObjs[vn] = trObjs;
    }
    doShowVsc(vn);                  // sofort anzeigen: ohne Ruckeln beim Update!
    vState[vn] = 3;

    // Warteschlange der VSC-Requests weiter abarbeiten
    if (vnRdIdx != vnWrIdx) {
        vn = vnQueue[vnRdIdx];
        if (++vnRdIdx >= MAX_VSC)
            vnRdIdx = 0;
        vscSendReq(vFilename[vn], vTileId[vn]);
        vState[vn] = 2;
    } else {
        bVscBusy = false;
        cntRetry = 0;
if (LOG>2) console.log("VSC-Queue abgearbeitet");
    }


    // Liste der Kameras, Symbole und Sensoren aller sichtbaren Tiles neu bauen,
    // der Einfachheit halber jedesmal komplett neu. Dabei globale Objekte
    // in das Zentralregister eintragen; als Text aufbereiten.

    if (tCamList !== undefined)
        delete tCamList;
    tCamList = new Array();
    if (tSymbList !== undefined)
        delete tSymbList;
    tSymbList = new Array();
    if (tSensList !== undefined)
        delete tSensList;
    tSensList = new Array();
    if (tDynaList !== undefined)
        delete tDynaList;
    tDynaList = new Array();
    var str = "";                       // globale Objekte sammeln
    var nCam = 0;                       // zur Anzeige
    var nSym = 0;
    var nSen = 0;
    var nDyn = 0;

    for (var n = 0; n < MAX_VSC; n++) { // fuer alle geladenen VSC-Scripts:
        if (vFilename[n] == "")
            continue;
        var tileId = vTileId[n];
        var tnr = idToNum(tileId);      // Nummer in Tile-Liste
        var cams = vCameras[n];         // Kameras = Viewpoints = Sprungziele
        if (cams !== undefined) {
            for (var i = 0; i < cams.length; i++) {
                var cam = new Array();
                cam[0] = cams[i][0];
                cam[1] = cams[i][1] + qXofs[tnr];
                cam[2] = cams[i][2];
                cam[3] = cams[i][3] + qZofs[tnr];
                cam[4] = cams[i][4];
                cam[5] = cams[i][5];
                cam[6] = cams[i][6];
                if (cams[i][7] == undefined)
                    cam[7] = 45.0;      // Angle ggf pauschal
                else
                    cam[7] = cams[i][7];
                tCamList.push(cam);
                // globale Kameras der akt. Tile veroeffentlichen
                if ((tileId == qid) && (cam[0].indexOf("g_") == 0))
                    str = str + "camera;"+cam[0]+";"+cam[1]+";"+cam[2]+";"+cam[3]+";"
                        +cam[4].toFixed(3)+";"+cam[5].toFixed(3)+";"+cam[6].toFixed(3)+";"+cam[7]+"|";
            }
            nCam += cams.length;
        }
        var symbols = vSymbols[n];      // Symbole
        if (symbols !== undefined) {
            for (var i = 0; i < symbols.length; i++) {
                var sym = symbols[i];
                tSymbList.push(sym);
                // globale Symbole der akt. Tile veroeffentlichen
                if ((tileId == qid) && (sym[0].indexOf("g_") == 0))
                    str = str + "symbol;"+sym[0]+";"+sym[1]+";"+sym[2]+";"
                        +sym[3]+";"+sym[4]+";"+sym[5]+"|";
            }
            nSym += symbols.length;
        }
        var sensors = vSensors[n];      // Sensoren
        if (sensors !== undefined) {
            for (var i = 0; i < sensors.length; i++) {
                var sen = new Array();
                sen[0] = sensors[i][0];
                sen[1] = sensors[i][1] + qXofs[tnr];
                sen[2] = sensors[i][2];
                sen[3] = sensors[i][3] + qZofs[tnr];
                sen[4] = sensors[i][4];
                sen[5] = sensors[i][5];
                sen[6] = sensors[i][6];
                tSensList.push(sen);
                // globale Sensoren der akt. Tile veroeffentlichen
                if ((tileId == qid) && (sen[0].indexOf("g_") == 0))
                    str = str + "sensor;"+sen[0]+";"+sen[1]+";"+sen[2]+";"+sen[3]+";"
                        +sen[4]+";"+sen[5]+";"+sen[6]+"|";
            }
            nSen += sensors.length;
        }
        var dynamics = vDynamics[n];    // Dynamische Verknuepfungen
        if (dynamics !== undefined) {
            for (var i = 0; i < dynamics.length; i++) {
                var dyn = dynamics[i];
                tDynaList.push(dyn);
            }
            nDyn += dynamics.length;
        }
    }

    if (str != "") {                    // gibt es globale Objekte?
        var xmlhttp = new XMLHttpRequest();
        xmlhttp.onreadystatechange = function() {
            if ((xmlhttp.readyState == 4) && (xmlhttp.status == 200)) {
                var rxText = xmlhttp.responseText;
                if (rxText.indexOf("*") == 0)
                    alert("xglobal.php: " + rxText);
            }
        };
        xmlhttp.open("GET", "../php/xglobal.php?set=w" + qid + "|" + str, true);
        xmlhttp.send();
    }
if (LOG>1) console.log("Listen: "+nCam+" Cams, "+nSym+" Syms, "+nSen+" Sens, "+nDyn+" Dyns");
}

//-----------------------------------------------------------------------------

function ReadGlobalSymbols() {
    var xmlhttp = new XMLHttpRequest();
    xmlhttp.onreadystatechange = function() {
        if ((xmlhttp.readyState == 4) && (xmlhttp.status == 200)) {
            GlobalList = xmlhttp.responseText;
        }
    };
    xmlhttp.open("GET", "../php/xglobal.php?get=1", true);
    xmlhttp.send();
}

//-----------------------------------------------------------------------------
// Position der BoundingBoxes absolut machen

function handleBBoxOffsets(tn, bbox) {
    for (var i = 0; i < bbox.length; i++) {
        var bb = bbox[i];
        if (bb === undefined)
            break;
        bb[0] += qXofs[tn];
        bb[2] += qZofs[tn];
    }
}

function registerBBoxes(qid, bbox, bbFrames) {
    var tn = idToNum(qid);
    handleBBoxOffsets(tn, bbox);
    qBBoxes[tn] = bbox;
    bbFrames.position.set(qXofs[tn], 0, qZofs[tn]);
    qBBoxFrames[tn] = bbFrames;
}

//-----------------------------------------------------------------------------
// Sichtbarkeit der Tiles steuern. Jede Sekunde einmal aufgerufen.

function manageSceneLoader() {

/*
    if (++loadState < 2)            // erstmal gaanz langsam
        return;
    loadState = 0;
*/
    if (setTileOffsets()) {
//        curIdx = idToNum(curTid);   //@@ weitere Aktionen an Kachelgrenzen ??
    }
    if (MyUserId <= 0)              // Init noch nicht fertig?
        return;                     // warten

    vscSendReq("", curTid);         // Aenderungen im VSC-Umfeld abfragen

    if (_DEBUG) {                   // Testbetrieb:
        doShowScene(curTid);        // nur die aktuelle Kachel beackern
        return;
    }

    // Distanzradien von eigener Position zu den Mittelpunkten der Kacheln
    var dst1 = ViewRange;           // Nachbar wird sichtbar, ggf. nachladen
    var dst2 = 1.2 * ViewRange;     // Nachbar wird unsichtbar
    var dst3 = 2.2 * ViewRange;     // Nachbar wird entfernt

    for (var x = ofsX-1; x <= ofsX+1; x++) {
        for (var z = ofsZ-1; z <= ofsZ+1; z++) {
            var qid = (x+50)*100 + (z+50);
            // Abstand zur Kachelmitte des Nachbarn
            var dx = posX - x * LEN_QUAD;
            var dz = posZ - z * LEN_QUAD;
            var dist = Math.sqrt(dx*dx + dz*dz);
            if (dist > dst3) {
                doTerminateScene(qid);
            } else if (dist > dst2) {
                doHideScene(qid);
            } else if ((dist < dst1) || (qid == curTid)) {
                if (doShowScene(qid))
                    break;          // falls neu gebaut: abwarten
            }
        }
    }
}

//-----------------------------------------------------------------------------

var getCurSceneVersion = function() {
    if (!bSceneValid || (qTiles[curIdx] === undefined))
        return "(no scene)";
    return qTiles[curIdx].qGetSceneVersion;
};

function testCurScenePosition(xp, yp, zp) {
    if (!bSceneValid)
        return true;
    if (!testBoundBoxFree(qBBoxes[curIdx], xp, yp, zp))
        return false;
    for (var vn = 0; vn < MAX_VSC; vn++) {
        if (vTileId[vn] == curTid) {
            if (!testBoundBoxFree(vBBoxes[vn], xp, yp, zp))
                return false;
        }
    }
    if (qTiles[curIdx] === undefined)
        return true;
    return qTiles[curIdx].qTestScenePosition(xp, yp, zp);
}

function getCurSceneGroundLevel(xp, yp, zp) {
    if (!bSceneValid)
        return 0;
    var lev = getBoundBoxGroundLevel(qBBoxes[curIdx], xp, yp, zp);
    if (lev != 0)
        return lev;
    for (var vn = 0; vn < MAX_VSC; vn++) {
        if (vTileId[vn] == curTid) {
            lev = getBoundBoxGroundLevel(vBBoxes[vn], xp, yp, zp);
            if (lev != 0)
                return lev;
        }
    }
    if (qTiles[curIdx] === undefined)
        return 0;
    return qTiles[curIdx].qGetSceneGroundLevel(xp, yp, zp);
}

function showBoundingBoxes(bShow) {
    for (var i = 0; i < MAX_TILES; i++) {
        if (qVisible[i])
            doBBShow(qBBoxFrames[i], bShow);
    }
    for (var i = 0; i < MAX_VSC; i++) {
        if (vVisible[i])
            doBBShow(vBBoxFrames[i], bShow);
    }
}

function doBBShow(bb, bShow) {
    if (bb !== undefined) {
        if (bShow)
            scene.add(bb);
        else
            scene.remove(bb);
    }
}

//-----------------------------------------------------------------------------
// Dynamische Operationen an allen sichtbaren Kacheln weiterfuehren

function cycleTiles(msPerFrame) {

    if (!bSceneValid)
        return;

    for (var tn = 0; tn < MAX_TILES; tn++) {
        if (!qVisible[tn])
            continue;
        qTiles[tn].qCycleScene(msPerFrame);
        doTileDynamics(msPerFrame);
        if (doTileSensors(msPerFrame))
            return;
    }
}

var findSymbol = function(name) {
    var bFound = false;
    var sym;
    for (var i = 0; i < tSymbList.length; i++) {
        sym = tSymbList[i];
        if (sym[0] == name) {
            bFound = true;
            break;
        }
    }
    if (bFound)
        return sym;
    return undefined;
};

var findCamera = function(name) {
    var bFound = false;
    var cam;
    for (var i = 0; i < tCamList.length; i++) {
        cam = tCamList[i];
        if (cam[0] == name) {
            bFound = true;
            break;
        }
    }
    // falls lokal nicht gefunden: global suchen
    if (!bFound && (GlobalList !== undefined)) {
        var pos = GlobalList.indexOf(name);
        if (pos > 0) {
            var ePos = GlobalList.indexOf("|", pos);
            var camStr = GlobalList.slice(pos, ePos);
            camItems = camStr.split(";");
            cam = Array(camItems[0], parseFloat(camItems[1]), parseFloat(camItems[2]),
                parseFloat(camItems[3]), parseFloat(camItems[4]), parseFloat(camItems[5]),
                parseFloat(camItems[6]), parseFloat(camItems[7]));
            bFound = true;
        }
    }
    if (bFound)
        return cam;
    return undefined;
};

//-----------------------------------------------------------------------------
// Dynamische Bewegungen im aktuellen Kontext bearbeiten

function doTileDynamics(msPerFrame) {

    if ((tSymbList === undefined) || (tDynaList === undefined))
        return;

    for (var i = 0; i < tDynaList.length; i++) {
        var dyn = tDynaList[i];     // 0: oName, 1: symName, 2: type, 3: curVal, 4: curTim
        var oName = dyn[0];
        var obj = scene.getObjectByName(oName);
        if (obj === undefined)      // Objekt nicht gefunden
            continue;
        var sym = findSymbol(dyn[1]);
        if (sym === undefined)
            continue;               // Symbol nicht gefunden

        // alles bereit: Bewegung je nach Typ ausfuehren
        var typ = dyn[2];
        var vCur = dyn[3];          // aktueller dynamischer Wert
        var tCur = dyn[4];          // aktuelle Zeitposition
        var vBeg = sym[1];          // Startwert
        var vEnd = sym[2];          // Endewert
        var dTim = sym[3]*1000;     // Zeit in ms
        var flgs = sym[4];          // Flags

        if (!(flgs & 0x8000)) {     // durch Sensor gesteuert?
            tCur += msPerFrame;
            if (tCur > dTim)
                tCur -= dTim;
            var vDif = (vEnd - vBeg) * msPerFrame / dTim;
            if (flgs & 0x01) {      // hin und zurueck
                if (tCur < dTim/2) {
                    vCur += 2*vDif;
                    if (vCur > vEnd)
                        vCur = vEnd;
                } else {
                    vCur -= 2*vDif;
                    if (vCur < vBeg)
                        vCur = vBeg;
                }
            } else {
                vCur += vDif;
                if (vCur >= vEnd)
                    vCur = vBeg;
            }
            dyn[3] = vCur;
            dyn[4] = tCur;
        } else
            vCur = sym[5];

        if (typ == "x0")
            obj.position.x = vCur;
        else if (typ == "y0")
            obj.position.y = vCur;
        else if (typ == "z0")
            obj.position.z = vCur;
        else if (typ == "rx")
            obj.rotation.x = vCur * Math.PI/180;
        else if (typ == "ry")
            obj.rotation.y = vCur * Math.PI/180;
        else if (typ == "rz")
            obj.rotation.z = vCur * Math.PI/180;
        else if (typ == "siz")
            obj.scale.set(vCur, vCur, vCur);
    }
}

//-----------------------------------------------------------------------------
// Sensoren im aktuellen Kontext bearbeiten

function doTileSensors(msPerFrame) {

    if ((tSymbList === undefined) || (tSensList === undefined) || (tCamList === undefined))
        return;

    for (var i = 0; i < tSensList.length; i++) {
        var sen = tSensList[i];
        var x = sen[1];                     // bereits absolute Koordinaten 
        var y = sen[2]; 
        var z = sen[3];
        var minDist = sen[4];
        var maxDist = sen[5];
        var symName = sen[6];
        var dx = posX - x;
        var dz = posZ - z;
        var curDist = Math.sqrt(dx*dx + dz*dz);
        if (curDist > maxDist)
            continue;                       // zu weit entfernt
        if ((posY < y - 2.4) || (posY > y + 2.4))
            continue;                       // andere Ebene @@

        var cam = findCamera(symName);
        var sym = findSymbol(symName);

        if (sym !== undefined) {
            if (curDist <= minDist) {       // ganz dicht dran?
                sym[5] = sym[2];            // curVal = endVal
                continue;
            }
            var senVar = maxDist - minDist; // Breite des Naeherungsbereichs
            if (senVar == 0) {
                sym[5] = sym[2];            // curVal = endVal
                continue;
            }
            var senCur = maxDist - curDist; // relative Position im Naeherungsbereich
            var symVar = sym[2] - sym[1];   // Werteband des Symbols
            sym[5] = sym[1] + symVar * senCur/senVar;
            continue;
        }

        if (curDist <= minDist) {           // dicht dran am Sensor?
            if (cam === undefined) {
                dbgMessage = "### Teleport-Empfaenger nicht bereit: "+symName;
            } else {
if (LOG) console.log("Teleport von "+posX.toFixed(1)+"/"+posZ.toFixed(1)+" nach "+cam[1]+"/"+cam[3]+" ("+cam[0]+")");
                posX = cam[1];              // Positionen sind absolut
                posY = cam[2];
                posZ = cam[3];
                rotX = cam[4] * 180/Math.PI;
                rotY = cam[5] * 180/Math.PI;
                rotZ = cam[6] * 180/Math.PI;
                velTouch = 0;               // Bewegung stoppen
                bUpKey = false;
                changeCamera(cam, 0);       // Sprung
                return true;                // Schluss hier
            }
            continue;
        }
    }
    return false;
}

//-----------------------------------------------------------------------------
// Zur naechsten Kameraposition wechseln

function gotoNextViewpoint(back) {

    if (!bSceneValid)
        return;
    if (back === undefined)
        back = false;

    var num = tCamList.length;
    var newCam = curCamera;
    if (back) {
        if (newCam > 0)
            newCam--;
        else
            newCam = num;
    } else {
        if (num > 0) {
            newCam++;
            if (newCam > num)
                newCam = 0;
        }
    }
    if (newCam == curCamera)
        return;

    var cam;
    if (newCam == 0)
        cam = new Array("FPV", posX, posY, posZ, rotX, rotY, rotZ);
    else
        cam = tCamList[newCam-1];
    changeCamera(cam, newCam);
}

