//--------------------------------------------------------------------------
// w3dTools.js - Utilities fuer 3D-Wiki-Szenerien               2016-02-19

// eine Art "Mensch-Aergere-Dich-Nicht"-Figur mit spitzer Nase, ca. 2m gross

var AvatarObj = function() {

    var avaGeo = new THREE.Geometry();
    var mat = new THREE.MeshLambertMaterial({color: 0xCCAA00,
           transparent:true, opacity:1.0, side: THREE.DoubleSide});

    // RadTop, RadBot, Height, RadSegm, HeiSegm, OpenEnded
    var geo = new THREE.CylinderGeometry(0.42, 0.42, 0.3, 10, 2, 0);
    var obj = new THREE.Mesh(geo, mat);
    obj.position.y = 0.15;
    obj.updateMatrix();
    avaGeo.merge(obj.geometry, obj.matrix);

    geo = new THREE.CylinderGeometry(0.13, 0.42, 1.2, 10, 4, 0);
    obj = new THREE.Mesh(geo, mat);
    obj.position.y = 0.9;
    obj.updateMatrix();
    avaGeo.merge(obj.geometry, obj.matrix);

    geo = new THREE.SphereGeometry(0.38, 8, 8);
    obj = new THREE.Mesh(geo, mat);
    obj.position.y = 1.85;
    obj.updateMatrix();
    avaGeo.merge(obj.geometry, obj.matrix);

    geo = new THREE.CylinderGeometry(0.01, 0.15, 0.32, 10, 2, 0);
    obj = new THREE.Mesh(geo, mat);
    obj.position.y = 1.83;
    obj.position.x = -0.50;
    obj.rotation.z = Math.PI/2;
    obj.updateMatrix();
    avaGeo.merge(obj.geometry, obj.matrix);

    return new THREE.Mesh(avaGeo, mat);
};

//-----------------------------------------------------------------------------

var Tanne = function() {

    var matGreen = new THREE.MeshLambertMaterial({color: 0x005F1F, side: THREE.DoubleSide});
    var matBrown = new THREE.MeshLambertMaterial({color: 0x5F3F0F, side: THREE.DoubleSide});

    var tanne = new THREE.Mesh();
    
    var geo = new THREE.CylinderGeometry(0.0, 0.4, 1.2, 6, 2, 0);
    var obj = new THREE.Mesh(geo, matGreen);
    obj.position.y = 4.1;
    tanne.add(obj);
    
    geo = new THREE.CylinderGeometry(0.1, 0.9, 1.5, 6, 2, 0);
    obj = new THREE.Mesh(geo, matGreen);
    obj.position.y = 3.0;
    tanne.add(obj);

    geo = new THREE.CylinderGeometry(0.2, 1.4, 1.5, 6, 2, 0);
    obj = new THREE.Mesh(geo, matGreen);
    obj.position.y = 2.0;
    tanne.add(obj);

    geo = new THREE.CylinderGeometry(0.4, 2.0, 1.5, 6, 2, 0);
    obj = new THREE.Mesh(geo, matGreen);
    obj.position.y = 1.0;
    tanne.add(obj);

    geo = new THREE.CylinderGeometry(0.2, 0.2, 1.0, 8, 2, 0);
    obj = new THREE.Mesh(geo, matBrown);
    obj.position.y = 0.5;
    tanne.add(obj);
    tanne.name = "TanneObj";

    return tanne;
};

//-----------------------------------------------------------------------------
// Farbiges Achsenkreuz aus RGB-Pfeilen

function setAchsenkreuz(siz) {

    matRed = new THREE.MeshLambertMaterial({color: 0xFF0000, side:THREE.DoubleSide});
    matGreen = new THREE.MeshLambertMaterial({color: 0x00FF00, side:THREE.DoubleSide});
    matBlue = new THREE.MeshLambertMaterial({color: 0x0000FF, side:THREE.DoubleSide});

    var achsGeo = new THREE.Geometry();

    var geo = new THREE.CubeGeometry(0.02, 1, 0.02);
    var obj = new THREE.Mesh(geo, matRed);
    obj.updateMatrix();
    achsGeo.merge(obj.geometry, obj.matrix);

    geo = new THREE.CylinderGeometry(0.0, 0.036, 0.1, 8, 2, 1);
    obj = new THREE.Mesh(geo, matRed);
    obj.position.y = 0.5;
    obj.updateMatrix();
    achsGeo.merge(obj.geometry, obj.matrix);

    var cross = new THREE.Mesh();

    var achsObj = new THREE.Mesh(achsGeo, matRed);
    achsObj.rotation.z = -Math.PI/2;
    achsObj.position.x = 0.5;
    cross.add(achsObj);

    achsObj = new THREE.Mesh(achsGeo, matGreen);
    achsObj.position.y = 0.5;
    cross.add(achsObj);

    achsObj = new THREE.Mesh(achsGeo, matBlue);
    achsObj.rotation.x = Math.PI/2;
    achsObj.position.z = 0.5;
    cross.add(achsObj);

    if (siz === undefined)
        siz = 1.0;
    cross.scale.set(siz, siz, siz);
    cross.name = "CrossObj";

    return cross;
}

//-------------------------------------------------------------------------
// Liniengitter in der Basisebene bauen
    
function buildGrid() {

    var Grid = new THREE.Mesh();
    var y = -0.30;
    var xz = 2*LEN_QUAD;
    for (var x = -xz; x <= xz; x += 10) {
        geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(x, y, -xz));
        geo.vertices.push(new THREE.Vector3(x, y, xz));
        if (x == 0)
            obj = new THREE.Line(geo, new THREE.LineBasicMaterial({color:0xFFFF40}));
        else
            obj = new THREE.Line(geo, new THREE.LineBasicMaterial({color:0x40FF40}));
        Grid.add(obj);
    }
    for (var z = -xz; z <= xz; z += 10) {
        geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(-xz, y, z));
        geo.vertices.push(new THREE.Vector3(xz, y, z));
        if (z == 0)
            obj = new THREE.Line(geo, new THREE.LineBasicMaterial({color:0xFFFF40}));
        else
            obj = new THREE.Line(geo, new THREE.LineBasicMaterial({color:0x40FF40}));
        Grid.add(obj);
    }
    return Grid;
}

//-------------------------------------------------------------------------

function buildGround(texFile, scale) {

    var groundObj = new THREE.Mesh();

    // Umrandung als Linie einzeichnen
    var x1 = LEN_QUAD/2;
    var h1 = 0.02;
    var geometry = new THREE.Geometry();
    geometry.vertices.push(new THREE.Vector3(-x1, h1, -x1));
    geometry.vertices.push(new THREE.Vector3( x1, h1, -x1));
    geometry.vertices.push(new THREE.Vector3( x1, h1,  x1));
    geometry.vertices.push(new THREE.Vector3(-x1, h1,  x1));
    geometry.vertices.push(new THREE.Vector3(-x1, h1, -x1));
    var line = new THREE.Line(geometry, new THREE.LineBasicMaterial({color:0x2FFF2F}));
    groundObj.add(line);

    // Achsenkreuz am Boden
    var x2 = LEN_QUAD/2;
    var mat = new THREE.LineBasicMaterial({color:0x9F9F1F});
    geometry = new THREE.Geometry();
    geometry.vertices.push(new THREE.Vector3(0, h1, x2));
    geometry.vertices.push(new THREE.Vector3( 0, h1, -x2));
    line = new THREE.Line(geometry, mat);
    groundObj.add(line);
    geometry = new THREE.Geometry();
    geometry.vertices.push(new THREE.Vector3(x2, h1, 0));
    geometry.vertices.push(new THREE.Vector3(-x2, h1, 0));
    line = new THREE.Line(geometry, mat);
    groundObj.add(line);


//    var gndWidth = LEN_QUAD * 1.5;
    var gndWidth = LEN_QUAD;

//THREE.ImageUtils.loadTexture is being deprecated. Use THREE.TextureLoader() instead.
//http://threejs.org/docs/#Reference/Loaders/TextureLoader
/*
  var maptex = THREE.ImageUtils.loadTexture(texFile);  #######
    maptex.wrapS = THREE.RepeatWrapping;
    maptex.wrapT = THREE.RepeatWrapping;
    maptex.repeat = new THREE.Vector2(gndWidth/rep, gndWidth/rep);
    maptex.anisotropy = renderer.getMaxAnisotropy();
    var matBoden = new THREE.MeshLambertMaterial({map:maptex, side:THREE.DoubleSide});
*/

var loader = new THREE.TextureLoader();  //// instantiate a loader (nur einmalig)
  loader.load(                             //// load a resource (mit callback wenn fertig!)
    texFile,                   // resource URL
      function ( texture ) {  // Function when resource is loaded (callback)      
      if(scale===undefined) scale = 4;
      texture.wrapS = THREE.RepeatWrapping;
      texture.wrapT = THREE.RepeatWrapping;
      texture.repeat = new THREE.Vector2(gndWidth/scale, gndWidth/scale);
        texture.anisotropy = 16;
//        texture.anisotropy = renderer.getMaxAnisotropy();
        var matBoden  = new THREE.MeshLambertMaterial({map:texture, side:THREE.DoubleSide});// do something with the texture
        var plane     = new THREE.Mesh(new THREE.PlaneGeometry(gndWidth, gndWidth, 4, 4), matBoden);
        plane.rotation.x = -Math.PI/2;
        plane.position.set(0, -0.01, 0);
        groundObj.add(plane); // !! Wird erst nach dem Laden asyncron ergänzt!
      } // end callback
  );// end .loader

    groundObj.name = "Ground";

    return groundObj;
}

//-----------------------------------------------------------------------------
// ein grobes Raster senkrechter Latten zur Orientierung in der Flaeche

function buildPins() {

    var geoSum = new THREE.Geometry();
    var geoPin = new THREE.CubeGeometry(0.12, 10, 0.12);
    var matPin = new THREE.MeshLambertMaterial({color: 0xbf7f00});
    for (var x = -3; x < 3; x++) {
        for (var z = -3; z < 3; z++) {
            var obj = new THREE.Mesh(geoPin, matPin);
            obj.position.set(x*20+10, 5, z*20+10);
            obj.updateMatrix();
            geoSum.merge(obj.geometry, obj.matrix);
        }
    }
    var pinsObj = new THREE.Mesh(geoSum, matPin);
    pinsObj.Name = "Pins";
    return pinsObj;
}

function testPinPosition(xp, yp, zp) {

    if ((yp < 0) || (yp > 10))
        return true;                    // darunter oder darueber: frei
    var dd = 0.3;
    for (var x = -3; x < 3; x++) {
        for (var z = -3; z < 3; z++) {
            var xx = x*20+10;
            var zz = z*20+10;
            if ((xp > xx-dd) && (xp < xx+dd) && (zp > zz-dd) && (zp < zz+dd))
                return false;            // Kollision
        }
    }
    return true;                        // keine Kollision
}


//-----------------------------------------------------------------------------
// Draw text on canvas

var CanvasTextObject = function(msg, fontName, fillStyle) {

    var cDim = 256;                 // sollte 2er-Potenz sein
    var canvas = document.createElement('canvas');
    canvas.width = cDim;
    canvas.height = cDim/2;
    var ctx = canvas.getContext('2d');

    ctx.fillStyle = 'red';
    ctx.fillRect(0, 0, canvas.width, canvas.height);
    ctx.fillStyle = 'white';
    ctx.fillRect(6, 6, canvas.width - 12, canvas.height - 12);
    ctx.font = fontName;
    ctx.fillStyle = fillStyle;
    ctx.textAlign = 'center';
    ctx.textBaseline = 'middle';
    ctx.fillText(msg, canvas.width/2, canvas.height/2);

    // canvas contents will be used for a texture
    var texture1 = new THREE.Texture(canvas);
    texture1.needsUpdate = true;

    var matCanv = new THREE.MeshBasicMaterial({map: texture1, side:THREE.DoubleSide});
    matCanv.transparent = true;

    var mesh1 = new THREE.Mesh(new THREE.PlaneGeometry(2.0, 1.0), matCanv);
//    var mesh1 = new THREE.Mesh(new THREE.CubeGeometry(2.0, 1.0, 0.02), matCanv);
    return mesh1;
};



//-----------------------------------------------------------------------------
// BoundingBox-Datensatz erzeugen

var makeBBox = function(x, y, z, wid, hei, len, oName) {
    return new Array(x, y, z, wid, hei, len, oName);
};

//-----------------------------------------------------------------------------

function showBoundBoxes(boxList) {

    var tol = 0.02;
    var mat = new THREE.LineBasicMaterial({color:0x009FFF});
    var boxObj = new THREE.Mesh();
    for (var i = 0; i < boxList.length; i++) {
        var rec = boxList[i];
        var x0 = rec[0];
        var y0 = rec[1];
        var z0 = rec[2];
        var wid = rec[3]/2+tol;
        var hei = rec[4]+tol;
        var len = rec[5]/2+tol;
        // Bodenflaeche
        var geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(x0-wid, y0, z0-len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0, z0-len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0, z0+len));
        geo.vertices.push(new THREE.Vector3(x0-wid, y0, z0+len));
        geo.vertices.push(new THREE.Vector3(x0-wid, y0, z0-len));
        // Deckflaeche
        geo.vertices.push(new THREE.Vector3(x0-wid, y0+hei, z0-len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0+hei, z0-len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0+hei, z0+len));
        geo.vertices.push(new THREE.Vector3(x0-wid, y0+hei, z0+len));
        geo.vertices.push(new THREE.Vector3(x0-wid, y0+hei, z0-len));
        boxObj.add(new THREE.Line(geo, mat));
        // noch drei Seitenlinien
        geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(x0+wid, y0, z0-len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0+hei, z0-len));
        boxObj.add(new THREE.Line(geo, mat));
        geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(x0+wid, y0, z0+len));
        geo.vertices.push(new THREE.Vector3(x0+wid, y0+hei, z0+len));
        boxObj.add(new THREE.Line(geo, mat));
        geo = new THREE.Geometry();
        geo.vertices.push(new THREE.Vector3(x0-wid, y0, z0+len));
        geo.vertices.push(new THREE.Vector3(x0-wid, y0+hei, z0+len));
        boxObj.add(new THREE.Line(geo, mat));
    }
    return boxObj;
}

var yMaxStufe = 0.4;
var hRaum = 2.4;

function testBoundBoxFree(boundBoxes, xp, yp, zp) {

    if (boundBoxes === undefined)
        return true;

    var hMax = yp + hRaum;
    for (var i = 0; i < boundBoxes.length; i++) {
        var rec = boundBoxes[i];
        var wid = rec[3]/2;
        var len = rec[5]/2;
        if ((xp < rec[0]-wid) || (xp > rec[0]+wid) || (zp < rec[2]-len) || (zp > rec[2]+len))
            continue;
        // innerhalb des x/z-Bereichs:
        var yLo = rec[1];
        if (yLo > hMax)
            continue;                   // Unterkante eine Etage hoeher
        var yHi = yLo + rec[4];         // Oberkante der Box
        if (yp < yHi - yMaxStufe)
            return false;
    }
    return true;
}

function getBoundBoxGroundLevel(boundBoxes, xp, yp, zp) {

    if (boundBoxes === undefined)
        return 0;

    var hMax = yp + hRaum;
    var hCur = -1001;               // auch Keller zulassen
    for (var i = 0; i < boundBoxes.length; i++) {
        var rec = boundBoxes[i];
        var x = rec[0];
        var y = rec[1];
        var z = rec[2];
        if (rec[6] != "") {         // bewegte Objekte symbolisch finden
            var obj = scene.getObjectByName(rec[6]);
            if (obj !== undefined) {
                x = obj.position.x;
                y = obj.position.y;
                z = obj.position.z;
            }
        }
        var wid = rec[3]/2;
        var len = rec[5]/2;
        if ((xp < x-wid) || (xp > x+wid) || (zp < z-len) || (zp > z+len))
            continue;
        // innerhalb des x/z-Bereichs:
        var yLo = y;
        var yHi = yLo + rec[4];
        if (yLo > hMax)
            continue;                // Unterkante eine Etage hoeher
        if (hCur < yHi)
            hCur = yHi;
    }
    if (hCur < -1000)               // keine einzige Box ?   
       hCur = 0;                    // Normalnull annehmen
    return hCur;
}


//-----------------------------------------------------------------------------

var Infos = function() {

    var startTime = Date.now();
    var prevTime = startTime;
    var ms = 0;
    var count = 0;

    var container = document.createElement('div');
    container.id = 'cInfo';
    container.style.position = 'absolute';
    container.style.cssText = 'width:900px;opacity:0.8;cursor:pointer';

    var vInfo = document.createElement('div');
    vInfo.id = 'vInfo';
    vInfo.style.cssText =
        'color:#ff0;font-family:Helvetica,Arial,sans-serif;font-size:15px;line-height:25px';
    container.appendChild(vInfo);
    vInfo.textContent = 'keine Anzeige?';

    return {
        REVISION: 1,
        domElement: container,
        begin: function() {
            startTime = Date.now();
        },

        end: function() {
            var time = Date.now();
            ms = time - startTime;
            return time;
        },

        update: function(msg) {
            startTime = this.end();
            vInfo.textContent = msg;
            count++;
        }
    };
};

//-----------------------------------------------------------------------------

var MessageLine = function(color, fsize) {

    var container = document.createElement('div');
    container.id = 'msgLine';
//    container.style.position = 'absolute';
//    container.style.left = '0px';
    container.style.cssText = 'width:600px;opacity:0.9;cursor:pointer';
    var vMessage = document.createElement('div');
    vMessage.id = 'vMessage';
    vMessage.style.cssText =
        'color:'+color+';font-family:Helvetica,Arial,sans-serif;font-size:'
        +fsize+'px;line-height:25px';
    container.appendChild(vMessage);
    vMessage.textContent = 'keine Message?';

    return {
        domElement: container,
        update: function(msg) {
            vMessage.textContent = msg;
        }
    };
};

//-----------------------------------------------------------------------------------

var getGrobDistanz = function(dx, dz) {
    return (Math.abs(dx) + Math.abs(dz)) / 2;
};

var getDistanz = function(dx, dz) {
    return Math.sqrt(dx*dx + dz*dz);
};

var getKursRad = function(dx, dz) {
    if (Math.abs(dz) < 0.000001)        // Division durch Null verhindern
        dz = 0.000001;
    var ku = Math.atan(dx / dz);
    if (dz < 0)
        ku -= Math.PI;
    if (ku < 0)
        ku += 2*Math.PI;
    return ku;
};

var getKurs = function(dx, dz) {
    return getKursRad(dx, dz) * 180/Math.PI;
};

var getPeilung = function(dx, dz, ownRot) {
    var wi = ownRot;
    if (wi > 180)
        wi -= 360;
    else if (wi < -180)
        wi += 360;
    var peil = -wi - getKurs(dx, dz);
    if (peil < -180)
        peil += 360;
    else if (peil > 180)
        peil -= 360;
    return peil;
};

//-----------------------------------------------------------------------------------

