//-----------------------------------------------------------------------------
// w3dVsc.js - Umsetzung eines VSC-Scripts in eine Three.js-Szene.

var VscBuilder = function() {   // als Objekt geklammert

var VersText = "V16.02.18";

var prObjects;                  // Liste der Prototypes
var prNames;                    // Name je Prototype
var prCount;                    // Instanz-Counter je Prototype

var vsName;                     // 0: Name der Szene
var vsVersion;                  // 1: Versionskennung
var grBase;                     // 2: Summen-Objekt
var vsBoxes;                    // 3: Liste der BoundingBoxes
var vsBBoxFrames;               // 4: Wireframes der Boxes
var vsSymbols;                  // 5: Liste der Symbole
var vsDynamics;                 // 6: Liste der Dynamik-Verknuepfungen
var vsSensors;                  // 7: Liste der Sensorpunkte
var vsCameras;                  // 8: Liste der Kamerapositionen

var vsTextures;                 // Liste der Texturen
var curTexture, curRepU, curRepV;
var vsStrings;                  // Liste der Textkonstanten

var protoName = "";
var grProto = 0;

var prXofs, prYofs, prZofs;
var prXrot, prYrot, prZrot;
var prXsiz, prYsiz, prZsiz;
var snapMode;    // 0-off, 1-up, 2-down, 3-left, 4-right, 5-forward, 6-back
var bSnapShift;
var bGenBox;
var bOpenEnded;

var errMsg = "";
var incName = "";
var lCnt;
var lCntSave;

var texLoader = new THREE.TextureLoader();  // instantiate a loader (nur einmalig)

//-----------------------------------------------------------------------------

// Bestandteile der Zeile isolieren, ueberfluessige Leerzeichen aussondern
function splitParams(line, maxParms) {
    var parms = new Array();
    var items = line.split(" ");
    var n = 0;
    for (var i = 0; (n < maxParms) && (i < 1000); i++) {
        if (items[i] == undefined)
            break;
        if (items[i] != "") {
            parms.push(items[i]);
            n++;
        }
    }
    return parms;
}

function addErrorMsg(str) {
    if (errMsg == "")                   // Startinfo bei erster Meldung
        errMsg = "Reading " + vsName + ":\r\n";
    errMsg = errMsg + incName + "#" + lCnt + ": " + str + "\r\n";
}

//-----------------------------------------------------------------------------
// Material besorgen: Textur oder unifarben, ggf. teiltransparent

function getMaterial(col, repU, repV) {
    var texId = -1;
    var cu = 0;
    var cv = 0;
    var opa = 1.0;
    var bTransp = false;
    var n = col.indexOf("/");          // evt. Opacity nach Schraegstrich
    if (n > 0) {
        opa = rdFloatPar(col.slice(n+1, col.length));
        if ((opa > 1.0) || (opa < 0))
            opa = 1.0;
        else
            bTransp = true;
        col = col.slice(0, n);
    }
    if (curTexture >= 0) {
        texId = curTexture;
        cu = curRepU;
        cv = curRepV;
    } else for (var i = 0; i < vsTextures.length; i++) {
        if (col == vsTextures[i][0]) {
            texId = i;
            if ((repU != undefined) && (repV != undefined) && (repU > 0) && (repV > 0)) {
                cu = repU;
                cv = repV;
            }
            break;
        }
    }
    if (texId < 0)                  // einfarbiges Material
        return new THREE.MeshLambertMaterial({color:parseInt(col, 16),
              transparent:bTransp, opacity:opa, side:THREE.DoubleSide});

    // Textur-Material:
    var map = THREE.ImageUtils.loadTexture(vsTextures[texId][1]);
    map.wrapS = map.wrapT = THREE.RepeatWrapping;
    if ((cu > 0) && (cv > 0))
        map.repeat = new THREE.Vector2(cu, cv);
    map.anisotropy = renderer.getMaxAnisotropy();
    return new THREE.MeshLambertMaterial({map: map,
          transparent:bTransp, opacity:opa, side: THREE.DoubleSide});
/*
    // load a resource (mit callback wenn fertig!)
    var mat = undefined;
    texLoader.load(vsTextures[texId][1],
        function(texture) {
            texture.wrapS = texture.wrapT = THREE.RepeatWrapping;
            if ((cu > 0) && (cv > 0))
                texture.repeat = new THREE.Vector2(cu, cv);
            texture.anisotropy = renderer.getMaxAnisotropy();
            mat = new THREE.MeshLambertMaterial({map: map,
                    transparent:bTransp, opacity:opa, side: THREE.DoubleSide});
        } // end callback
    );// end texLoader

    //@@ und jetzt? Kann hier nicht warten...
*/
}

function addToGroup(geo, x, y, z, dx, dy, dz, col, repU, repV) {
    if (snapMode == 1)
        y += dy*prYsiz;
    else if (snapMode == 2)
        y -= dy*prYsiz;
    else if (snapMode == 3)
        x += dx*prXsiz;
    else if (snapMode == 4)
        x -= dx*prXsiz;
    else if (snapMode == 5)
        z -= dz*prZsiz;
    else if (snapMode == 6)
        z += dz*prZsiz;
    var obj = new THREE.Mesh(geo, getMaterial(col, repU, repV));
    obj.scale.set(prXsiz, prYsiz, prZsiz);
    obj.rotation.set(prXrot, prYrot, prZrot);
    obj.position.set(x+prXofs, y+prYofs, z+prZofs);
    if (protoName == "")
        grBase.add(obj);
    else
        grProto.add(obj);
}

function findPrototype(name) {
    for (var i = 0; i < prNames.length; i++) {
        if (name == prNames[i])
            return i;
    }
    return -1;
}

function findString(name) {
    for (var i = 0; i < vsStrings.length; i++) {
        var str = vsStrings[i];
        var n = str.indexOf(" ");
        if ((n >= 0) && (name == str.slice(0, n)))
            return (str.slice(n+1));
    }
    return "";
}

function addBBox(x, y, z, wid, hei, len, oName) {
    if (!bGenBox)
        return;
    if (oName === undefined)
        oName = "";
    var bb;
    if (protoName == "")
        bb = new makeBBox(x+prXofs, y+prYofs, z+prZofs, wid, hei, len, oName);
    else
        bb = new makeBBox(0, 0, 0, wid, hei, len, protoName);
    vsBoxes.push(bb);
}

function snapShift(dx, dy, dz) {
    if (!bSnapShift)
        return;
    if (snapMode == 1)
        prYofs += dy*prYsiz;
    else if (snapMode == 2)
        prYofs -= dy*prYsiz;
    else if (snapMode == 3)
        prXofs += dx*prXsiz;
    else if (snapMode == 4)
        prXofs -= dx*prXsiz;
    else if (snapMode == 5)
        prZofs -= dz*prZsiz;
    else if (snapMode == 6)
        prZofs += dz*prZsiz;
}

//-----------------------------------------------------------------------------
// TODO
function buildModel(logName, objFile, mtlFile) {
    // das Laden dauert länger, daher erstmal Platz reservieren
    prObjects.push(0);              // spaeter nachreichen
    prNames.push(logName);
    prCount.push(0);                // Instanzcounter noch Null

    var onProgress = function(xhr) { };
    var onError = function(xhr) {
        addErrorMsg("Model "+logName+" failed");
    };

    THREE.Loader.Handlers.add(/\.dds$/i, new THREE.DDSLoader());
    var loader = new THREE.OBJMTLLoader();
    loader.load(objFile, mtlFile, function(object) {
        // fertig geladen: Objekt in der Liste nachtragen
        for (var i = 0; i < prNames.length; i++) {
            if (prNames[i] == logName) {
                prObjects[i] = object;
                break;
            }
        }
    }, onProgress, onError);
}

//-----------------------------------------------------------------------------
// Dynamik-Tabelle: 0: objName, 1: symName, 2: type, 3: curVal, 4: curTim

var parseSymbParam = function(parm, type) {
    for (var i = 0; i < vsSymbols.length; i++) {
        var sym = vsSymbols[i];
        if (parm == sym[0]) {       // Symbolname
            vsDynamics.push(new Array("@@NAME@@", parm, type, sym[1], 0));
            return sym[1];          // Startwert
        }
    }
    return rdFloatPar(parm);        // kein Symbol: Zahlenwert
};

function provideObjectName(name) {
    for (var i = 0; i < vsDynamics.length; i++) {
        if (vsDynamics[i][0] == "@@NAME@@")
            vsDynamics[i][0] = name;
    }
}

//-----------------------------------------------------------------------------
// Float-Parameter auswerten; Arithmetik in runden Klammern wird durch eine
// eval()-Anweisung ausgerechnet. Defines muessen zu diesem Zeitpunkt bereits
// ersetzt sein. Rechnen mit symbolischen Variablen ist nicht moeglich.

var rdFloatPar = function(parm) {
    if (parm.indexOf("(") == 0) {   // Arithmetik on Klammern?
        if (parm.slice(parm.length-1) != ")") {
            addErrorMsg("Arithmetik: schließende Klammer fehlt");
            return parseFloat(parm.slice(1));
        }
        return eval(parm.slice(1, parm.length-1));
    }
    return parseFloat(parm);
};

//-----------------------------------------------------------------------------
// Dezimal-Integer-Parameter auswerten; Arithmetik in runden Klammern wird
// durch eine eval()-Anweisung ausgerechnet.

var rdIntPar = function(parm) {
    if (parm.indexOf("(") == 0) {   // Arithmetik on Klammern?
        if (parm.slice(parm.length-1) != ")") {
            addErrorMsg("Arithmetik: schließende Klammer fehlt");
            return parseInt(parm.slice(1));
        }
        return eval(parm.slice(1, parm.length-1));
    }
    return parseInt(parm);
};

//-----------------------------------------------------------------------------
// Testen ob ein Zeichen Buchstabe, Zahl oder Underscore ist.

var isAlphanum = function(chr) {
    if (chr.match(/\w/g))           // 0-9, a-z, A-Z oder _ ?
        return true;                // alphanumerisches Zeichen
    return false;                   // Trennzeichen
};


//=============================================================================
// ab hier Memberfunktionen des VscBuilder-Objekts

 return {
  version: VersText,

//-----------------------------------------------------------------------------
// Compiler fuer VSC-Skript

  doBuildVscScene: function(sName, vscText) {

    var scriptOrg = vscText;
    vsName = sName;
    vsVersion = "";

    prXofs = prYofs = prZofs = 0;
    prXrot = prYrot = prZrot = 0;
    prXsiz = prYsiz = prZsiz = 1.0;
    snapMode = 0;
    bSnapShift = false;
    bGenBox = true;
    bOpenEnded = false;

    var prXofs_S, prYofs_S, prZofs_S;
    var prXrot_S, prYrot_S, prZrot_S;
    var prXsiz_S, prYsiz_S, prZsiz_S;
    var snapMode_S, bSnapShift_S;
        
    prObjects = new Array();
    prNames = new Array();
    prCount = new Array();    
    vsBoxes = new Array();    
    vsSymbols = new Array();
    vsDynamics = new Array();
    vsSensors = new Array();
    vsCameras = new Array();

    vsTextures = new Array();
    curTexture = -1;
    curRepU = curRepV = 0;
    vsStrings = new Array();

    var vsTracks = new Array();
    var bTrack = false;

    grBase = new THREE.Mesh();
    grProto = 0;
    protoName = "";
    var script = scriptOrg;
    var bComment = false;
    errMsg = "";
    lCnt = 0;
    lCntSave = 0;
    incName = "";
    var repCount = 0;
    var repText = "";


    // define-Liste anlegen und globale Definitionen vorbelegen
    var defNames = new Array();
    defNames.push("LEN_QUAD " + LEN_QUAD);
    defNames.push("FAR_PLANE " + FAR_PLANE);

    while (1) {                         // bis zum Ende des Skripts oder Fehler
        if (script.length < 3)
            break;

        var line = "";
        while (1) {                     // vollstaendige Zeile besorgen
            var n = script.indexOf("\n");
            if (n < 0)                  // letztes Newline fehlt?
                n = script.length;      // dann Restlaenge
            lCnt++;
            var cLine = script.slice(0, n);
            var pos = cLine.indexOf("string ");
            if (pos >= 0) {
                cLine = cLine.slice(pos+7);
                vsStrings.push(cLine.trim());
                script = script.slice(n+1, script.length);
                break;
            }
            // Zeilenkommentar rechts abschneiden
            var pos = cLine.indexOf("//");
            if (pos > 0)
                cLine = cLine.slice(0, pos);
            else {
                pos = cLine.indexOf("#");
                if (pos > 0)
                    cLine = cLine.slice(0, pos);
                else {
                    pos = cLine.indexOf(";");
                    if (pos > 0)
                        cLine = cLine.slice(0, pos);
                }
            }
            line = line + cLine;        // Zeilenanfang vorerst belassen
            line = line.trim();         // die zusammengefuegte trimmen
            script = script.slice(n+1, script.length);
            var lLen = line.length;
            if ((lLen > 0) && (line.charAt(lLen-1) == '\\'))
                line = line.slice(0, lLen-1) + ' ';
            else
                break;
        }

        line = line.trim();
        if (line.length < 2)
            continue;
        if (bComment) {
            if (line.indexOf("*/") >= 0)
                bComment = false;
            continue;
        }

        // Ende einer von xreadvsc.php eingefuegten Include-Passage erfassen
        if (line.indexOf("//@@ENDINC") == 0) {
            incName = "";
            lCnt = lCntSave;
            continue;
        }

        if ((line.charAt(0) == '#') || (line.charAt(0) == ';')
            || (line.substr(0, 2) == "//"))
            continue;
        if (line.substr(0, 2) == "/*") {
            bComment = true;
            continue;
        }
        if (line.substr(0, 6) == "endvsc")
            break;                      // Ende des VSC-Skripts

        // Gueltige VSC-Anweisung liegt jetzt vor:
        // Kommas und Tabs durch Blanks ersetzen
        var line = line.replace(/\,|\t/g, ' ');
        dLine = line.trim();

        //---------------------------------------------------------------------
        // In einer define-Zeile die ersten beiden Worte NICHT ersetzen.

        if (dLine.substr(0, 6) == "define") {       // "define HEIGHT  47.11"
            dLine = dLine.slice(7);
            dLine = dLine.trim();                   // "HEIGHT  47.11"
            var pos1 = dLine.indexOf(" ");
            if (pos1 < 0) {
                addErrorMsg("Leere Definition: " + dLine);
                break;
            }
            var defNam = dLine.slice(0, pos1);      // "HEIGHT"
            var defVal = dLine.slice(pos1+1);
            defVal = defVal.trim();                 // "47.11"

            var i;
            for (i = 0; i < defNames.length; i++) {
                var define = defNames[i];
                var posVal = define.indexOf(" ");
                if (posVal < 0)         // Define ohne Wert
                    continue;
                var strKey = define.slice(0, posVal);
                if (defNam == strKey) {
                    addErrorMsg(defNam + " ist bereits definiert");
                    break;
                }
            }
            if (i < defNames.length)
                break;
            defNames.push(dLine);       // ganze Zeile in die Liste
            continue;
        }

        //---------------------------------------------------------------------
        // Vor der Weiterverarbeitung der Zeile "defines" einsetzen

        /*** Karl ... ***/
        
        // A) So lange suchen, bis nichts mehr gefunden wird
        // B) DARIN alle defines suchen

        //console.log("+++", dLine)
        var bFound;
        do { // So lange suchen, bis nichts (mehr) gefunden wird
            bFound = false;
            for (var i = 0; i < defNames.length; i++) {  // Alle defines suchen
                // Define-Name und -Wert aus Liste lesen
                var define = defNames[i];   // ganze define-Zeile
                var posVal = define.indexOf(" ");
                if (posVal < 0)             // Define ohne Wert  ??? schon vorher nicht in Liste
                    continue;
                var strKey = define.slice(0, posVal);
                var strVal = define.slice(posVal+1, define.length);
                strVal = strVal.trim();
                // Define in Zeile suchen
                var posDef = dLine.indexOf(strKey);
                if (posDef < 0)     // dieses Define gibts hier nicht (mehr)
                    continue;       // Nächstes Define in dieser Zeile suchen
                // gefundenen Define-String pruefen
                if (posDef > 0) {
                    if (isAlphanum(dLine.charAt(posDef-1)))
                        continue;   // vor der Fundstelle kein Trennzeichen: Define doch nicht gefunden
                }
                var posEnd = posDef + strKey.length;
                if (posEnd < dLine.length) {
                    if (isAlphanum(dLine.charAt(posEnd)))
                        continue;   // hinter der Fundstelle kein Trennzeichen
                }
                // Define ist echt: Wert statt Name in Zeile einbauen
                var strRest = dLine.substr(posEnd, dLine.length);
                dLine = dLine.substr(0, posDef) + strVal + strRest;
                bFound = true; // Define gefunden. Es könnte (jetzt) noch mehr Defines geben,
                               //            ganze Zeile für ALLE Defines noch mal durchsuchen.
                               // Dadurch werden alle weiteren Defines gleichen oder aneren Namens gefunden.
                //console.log("   ", dLine)
            }
        } while (bFound); // Solange noch was gefunden wird, weiter versuchen
        

        /*** Martin ... *** /

        for (var i = 0; i < defNames.length; i++) {
            var define = defNames[i];   // ganze define-Zeile
            var posVal = define.indexOf(" ");
            if (posVal < 0)             // Define ohne Wert
                continue;
            var strKey = define.slice(0, posVal);
            var strVal = define.slice(posVal+1, define.length);
            strVal = strVal.trim();
            while (1) {
                var posDef = dLine.indexOf(strKey);
                if (posDef < 0)         // dieses Define gibts hier nicht (mehr)
                    break;
                var bCont = false;
                if (posDef > 0) {
                    if (isAlphanum(dLine.charAt(posDef-1)))
                        bCont = true;   // vor der Fundstelle kein Trennzeichen
                }
                var posEnd = posDef + strKey.length;
                if (posEnd < dLine.length) {
                    if (isAlphanum(dLine.charAt(posEnd)))
                        bCont = true;   // hinter der Fundstelle kein Trennzeichen
                }
                if (bCont) {            // hinter dem "Teilmatch" fortsetzen
                    // Probleme mit der Orion!
//                    dLine = dLine.slice(posEnd, dLine.length);
//                    if (dLine.length > 1)
//                        continue;
                    break;
                }
                var strRest = dLine.substr(posEnd, dLine.length);
                var newLine = dLine.substr(0, posDef) + strVal + strRest;
                dLine = newLine;
            }
        }
        / *** ... Martin ***/

        line = dLine;

        // haeufig verwendete lokale Variablen
        var parms, x0, y0, z0, wid, hei, len, rad, r1, r2, nsid, col;
        var rx, ry, rz, bx, by, bz, dx, dy, dz, arc, repU, repV;
        var geo, mat, obj;

        /*---------------------------------------------------------------------
         * include-Anweisung wertet der VSC-Lader auf Server-Seite aus und
         * expandiert das Script; hier werden nur die Zeilennummern verfolgt.
         * Schachtelungen sind verboten, weil zu komplex und auch unnoetig.
         */
        if (line.substr(0, 7) == "include") {
            if (line.indexOf("//@@") > 0) {
                addErrorMsg(line);      // Fehlermeldung durchreichen
                break;
            }
            incName = line.slice(8);
            lCntSave = lCnt;
            lCnt = 0;
            continue;
        }

        if (line.substr(0, 7) == "version") {
            parms = splitParams(line.slice(8), 2);
            vsVersion = parms[0];
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "repeat") {
            if (repCount != 0) {
                addErrorMsg("Keine geschachtelten Schleifen erlaubt");
                break;
            }
            parms = splitParams(line.slice(7), 2);
            repCount = rdIntPar(parms[0]);
            repText = script;           // Restscript lokal sichern
            continue;
        }

        if (line.substr(0, 6) == "endrep") {
            if (repCount > 0) {
                repCount--;
                script = repText;       // mit gesichertem Restscript fortsetzen
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 9) == "prototype") {
            if (protoName != "") {
                addErrorMsg("prototype geschachtelt in " + protoName);
                break;
            }
            parms = splitParams(line.slice(10), 1);
            var name = parms[0];
            var nobj = findPrototype(name);
            if (findPrototype(name) >= 0) {
                addErrorMsg("ein Prototype "+name+" exitiert bereits");
                break;
            }
            grProto = new THREE.Mesh();
            protoName = name;
            // globale Verschiebungsparameter zwischenspeichern
            prXofs_S = prXofs;
            prYofs_S = prYofs;
            prZofs_S = prZofs;
            prXrot_S = prXrot;
            prYrot_S = prYrot;
            prZrot_S = prZrot;
            prXsiz_S = prXsiz;
            prYsiz_S = prYsiz;
            prZsiz_S = prZsiz;
            snapMode_S = snapMode;
            bSnapShift_S = bSnapShift;
            // lokale Parameter neu beginnen
            prXofs = prYofs = prZofs = 0;
            prXrot = prYrot = prZrot = 0;
            prXsiz = prYsiz = prZsiz = 1.0;
            continue;
        }

        if (line.substr(0, 8) == "endproto") {
            if (protoName == "") {
                addErrorMsg("endproto unerwartet: kein prototype zuvor");
                break;
            }
            parms = splitParams(line.slice(8), 3);
            if ((parms.length != 0) && (parms.length != 3)) {
                addErrorMsg("endproto: ung&uuml;ltige Anzahl Parameter");
                break;
            }
            if (parms.length == 3) {
                wid = rdFloatPar(parms[0]);
                hei = rdFloatPar(parms[1]);
                len = rdFloatPar(parms[2]);
                addBBox(0, 0, 0, wid, hei, len, protoName);
            }
            grProto.name = protoName;
            prObjects.push(grProto);
            prNames.push(protoName);
            prCount.push(0);            // Instanzcounter noch Null
            protoName = "";
            // globale Verschiebungsparameter wieder herstellen
            prXofs = prXofs_S;
            prYofs = prYofs_S;
            prZofs = prZofs_S;
            prXrot = prXrot_S;
            prYrot = prYrot_S;
            prZrot = prZrot_S;
            prXsiz = prXsiz_S;
            prYsiz = prYsiz_S;
            prZsiz = prZsiz_S;
            snapMode = snapMode_S;
            bSnapShift = bSnapShift_S;
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "model") {
            parms = splitParams(line.slice(6), 3);
            if (parms.length < 3) {
                addErrorMsg("model: Zu wenige Parameter");
                break;
            }
            buildModel(parms[0], parms[1], parms[2]);
            continue;
        }

        if (line.substr(0, 7) == "texfile") {
            parms = splitParams(line.slice(8), 2);
            if (parms.length < 2) {
                addErrorMsg("texfile: Zu wenige Parameter");
                break;
            }
            var texdef = [parms[0], parms[1]];
            vsTextures.push(texdef);
            continue;
        }

        if (line.substr(0, 7) == "texture") {
            parms = splitParams(line.slice(8), 3);
            curTexture = -1;
            for (var i = 0; i < vsTextures.length; i++) {
                if (parms[0] == vsTextures[i][0]) {
                    curTexture = i;
                    break;
                }
            }
            if (curTexture == -1) {
                addErrorMsg("Textur " + parms[0] + " ist nicht definiert");
                break;
            }
            curRepU = curRepV = 0;
            if (parms.length > 1) {
                curRepU = rdFloatPar(parms[1]);
                if (parms.length > 2)
                    curRepV = rdFloatPar(parms[2]);
                else
                    curRepV = curRepU;
            }
            continue;
        }

        if (line.substr(0, 6) == "endtex") {
            if (curTexture == -1) {
                addErrorMsg("endtex unerwartet: keine texture-Passage hier");
                break;
            }
            curTexture = -1;
            curRepU = curRepV = 0;
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 9) == "translate") {
            parms = splitParams(line.slice(10), 3);
            if (parms.length < 3) {
                addErrorMsg("translate: Zu wenige Parameter");
                break;
            }
            prXofs = rdFloatPar(parms[0]);
            prYofs = rdFloatPar(parms[1]);
            prZofs = rdFloatPar(parms[2]);
            continue;
        }

        if (line.substr(0, 8) == "transrel") {
            parms = splitParams(line.slice(9), 3);
            if (parms.length < 3) {
                addErrorMsg("transrel: Zu wenige Parameter");
                break;
            }
            prXofs += rdFloatPar(parms[0]);
            prYofs += rdFloatPar(parms[1]);
            prZofs += rdFloatPar(parms[2]);
            continue;
        }

        if (line.substr(0, 6) == "rotate") {
            parms = splitParams(line.slice(7), 3);
            if (parms.length < 3) {
                addErrorMsg("rotate: Zu wenige Parameter");
                break;
            }
            prXrot = rdFloatPar(parms[0]) * Math.PI/180;
            prYrot = rdFloatPar(parms[1]) * Math.PI/180;
            prZrot = rdFloatPar(parms[2]) * Math.PI/180;
            continue;
        }

        if (line.substr(0, 6) == "rotrel") {
            parms = splitParams(line.slice(7), 3);
            if (parms.length < 3) {
                addErrorMsg("rotrel: Zu wenige Parameter");
                break;
            }
            prXrot += rdFloatPar(parms[0]) * Math.PI/180;
            prYrot += rdFloatPar(parms[1]) * Math.PI/180;
            prZrot += rdFloatPar(parms[2]) * Math.PI/180;
            continue;
        }

        if (line.substr(0, 6) == "scale ") {
            parms = splitParams(line.slice(6), 3);
            if ((parms.length != 1) && (parms.length != 3)) {
                addErrorMsg("scale: entweder 1 oder 3 Parameter");
                break;
            }
            prXsiz = rdFloatPar(parms[0]);
            if (parms.length == 1) {
                prYsiz = prXsiz;
                prZsiz = prXsiz;
            } else {
                prYsiz = rdFloatPar(parms[1]);
                prZsiz = rdFloatPar(parms[2]);
            }
            if (prXsiz * prYsiz * prZsiz == 0) {
                addErrorMsg("scale: Faktoren duerfen nicht Null sein");
                break;
            }
            continue;
        }

        if (line.substr(0, 8) == "scalerel") {
            parms = splitParams(line.slice(9), 3);
            if ((parms.length != 1) && (parms.length != 3)) {
                addErrorMsg("scalerel: entweder 1 oder 3 Parameter");
                break;
            }
            var fkt = rdFloatPar(parms[0]);
            prXsiz *= fkt;
            if (parms.length == 1) {
                prYsiz *= fkt;
                prZsiz *= fkt;
            } else {
                prYsiz *= rdFloatPar(parms[1]);
                prZsiz *= rdFloatPar(parms[2]);
            }
            if (prXsiz * prYsiz * prZsiz == 0) {
                addErrorMsg("scalerel: Faktoren duerfen nicht Null sein");
                break;
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 4) == "snap") {
            parms = splitParams(line.slice(5), 2);
            var mode = parms[0];
            if ((mode == "off") || (parseInt(mode) == 0))
                snapMode = 0;
            else if ((mode == "up") || (mode == "top") || (parseInt(mode) == 1))
                snapMode = 1;
            else if ((mode == "down") || (mode == "bottom") || (parseInt(mode) == 2))
                snapMode = 2;
            else if ((mode == "left") || (parseInt(mode) == 3))
                snapMode = 3;
            else if ((mode == "right") || (parseInt(mode) == 4))
                snapMode = 4;
            else if ((mode == "forward") || (parseInt(mode) == 5))
                snapMode = 5;
            else if ((mode == "backward") || (mode == "back") || (parseInt(mode) == 6))
                snapMode = 6;
            else {
                addErrorMsg("Ungueltiger snapMode: " + mode);
                break;
            }
            bSnapShift = false;
            if (parms[1] == "shift")
                bSnapShift = true;
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 3) == "set") {
            parms = splitParams(line.slice(4), 3);
            var swit;
            if ((parms[1] == "0") || (parms[1] == "off") || (parms[1] == "false"))
                swit = false;
            else if ((parms[1] == "1") || (parms[1] == "on") || (parms[1] == "true"))
                swit = true;
            else {
                addErrorMsg("Syntaxfehler Parameter 2");
                break;
            }
            if (parms[0] == "boxes")
                bGenBox = swit;
            else if (parms[0] == "openended")
                bOpenEnded = swit;
            else {
                addErrorMsg("Unbekannter Schalter: " + parms[0]);
                break;
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "symbol") {
            // symbol SymName vStart vEnd dTim  [flags]
            parms = splitParams(line.slice(7), 5);
            if (parms.length < 4) {
                addErrorMsg("symbol: Zu wenige Parameter");
                break;
            }
            var name = parms[0];
            var vStart = rdFloatPar(parms[1]);
            var vEnd = rdFloatPar(parms[2]);
            var dTim = rdFloatPar(parms[3]);
            var flags = 0;
            if (parms.length > 4)
                flags = parseInt(parms[4], 16);
            for (var i = 0; i < vsSymbols.length; i++) {
                var sym = vsSymbols[i];
                if (sym[0] == name) {
                    addErrorMsg("Symbol doppelt definiert: " + name);
                    break;
                }
            }
            vsSymbols.push(new Array(name, vStart, vEnd, dTim, flags, vStart));
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "sensor") {
            parms = splitParams(line.slice(7), 7);
            if (parms.length < 6) {
                addErrorMsg("sensor: Zu wenige Parameter");
                break;
            }
            var name = parms[0];
            x0 = rdFloatPar(parms[1]);
            y0 = rdFloatPar(parms[2]);
            z0 = rdFloatPar(parms[3]);
            var minDist = rdFloatPar(parms[4]);
            var maxDist = rdFloatPar(parms[5]);
            var symName = parms[6];
            var i;
            for (i = 0; i < vsSensors.length; i++) {
                var sen = vsSensors[i];
                if (sen[0] == name) {
                    addErrorMsg("Sensor-Name doppelt definiert: " + name);
                    break;
                }
            }
            for (i = 0; i < vsSymbols.length; i++) {
                var sym = vsSymbols[i];
                if (sym[0] == symName) {
                    sym[4] |= 0x8000;   // Automatik aussetzen
                    break;
                }
            }
            vsSensors.push(new Array(name, x0, y0, z0, minDist, maxDist, symName));
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "camera") {
            parms = splitParams(line.slice(7), 8);
            if (parms.length < 7) {
                addErrorMsg("camera: Zu wenige Parameter");
                break;
            }
            var name = parms[0];
            x0 = rdFloatPar(parms[1]);
            y0 = rdFloatPar(parms[2]);
            z0 = rdFloatPar(parms[3]);
            xr = rdFloatPar(parms[4]) * Math.PI/180;
            yr = rdFloatPar(parms[5]) * Math.PI/180;
            zr = rdFloatPar(parms[6]) * Math.PI/180;
            var ang = 45;
            if (parms.length > 7)
                ang = rdFloatPar(parms[7]);
            vsCameras.push(new Array(name, x0, y0, z0, xr, yr, zr, ang));
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "object") {
            parms = splitParams(line.slice(7), 12);
            var nPar = parms.length;
            if (nPar < 5) {
                addErrorMsg("object: Zu wenige Parameter");
                break;
            }
            var name = parms[0];
            var nobj = findPrototype(name);
            if (nobj < 0) {
                addErrorMsg("Kein Prototyp oder Model: " + name);
                break;
            }
            var siz = parseSymbParam(parms[1], "siz");
            x0 = parseSymbParam(parms[2], "x0");
            y0 = parseSymbParam(parms[3], "y0");
            z0 = parseSymbParam(parms[4], "z0");
            rx = ry = rz = 0;
            bx = by = bz = 0;
            if (nPar > 5) {                 // Rotation optional
                rx = parseSymbParam(parms[5], "rx") * Math.PI/180;
                ry = parseSymbParam(parms[6], "ry") * Math.PI/180;
                rz = parseSymbParam(parms[7], "rz") * Math.PI/180;
            }
            var useBox = false;
            if (nPar > 8) {
                bx = rdFloatPar(parms[8]);  // Bounding Box optional
                by = rdFloatPar(parms[9]);
                bz = rdFloatPar(parms[10]);
                if ((bx != 0) && (by != 0) && (bz != 0))
                    useBox = true;
            }
            var oName = prNames[nobj] + "_" + prCount[nobj];
            prCount[nobj]++;
            if (nPar > 11) {
                var p11 = parms[11]; 
                if (p11.charAt(0) == '\"')
                    oName = p11.slice(1, p11.length-1);
                else
                    var flg = parseInt(parms[11], 16);
                // TODO Flags zum Spiegeln usw.
            }
            if (useBox) {
                if (snapMode == 1)
                    y0 += by/2*prYsiz;
                else if (snapMode == 2)
                    y0 -= by/2*prYsiz;
                else if (snapMode == 3)
                    x0 += bx/2*prXsiz;
                else if (snapMode == 4)
                    x0 -= bx/2*prXsiz;
                else if (snapMode == 5)
                    z0 -= bz/2*prZsiz;
                else if (snapMode == 6)
                    z0 += bz/2*prZsiz;
            }
            var exemplar = prObjects[nobj].clone();
            exemplar.name = oName;
            exemplar.scale.set(siz*prXsiz, siz*prYsiz, siz*prZsiz);
            exemplar.rotation.set(rx+prXrot, ry+prYrot, rz+prZrot);
            exemplar.position.set(x0+prXofs, y0+prYofs, z0+prZofs);
            if (useBox) {
                addBBox(x0, y0, z0, siz*bx, siz*by, siz*bz, oName);
                snapShift(bx, by, bz);
            }
            if (protoName == "")
                grBase.add(exemplar);
            else
                grProto.add(exemplar);
            provideObjectName(oName);   // Namen in Dynamik-Liste ergaenzen
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "plane") {
            parms = splitParams(line.slice(6), 8);
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            wid = rdFloatPar(parms[3]);
            len = rdFloatPar(parms[4]);
            col = parms[5];
            var repU = 0;
            var repV = 0;
            if (parms.length > 6) {
                repU = rdFloatPar(parms[6]);
                if (parms.length > 7)
                    repV = rdFloatPar(parms[7]);
                else
                    repV = repU;
            }
            geo = new THREE.PlaneGeometry(wid, len, 4);
            obj = new THREE.Mesh(geo, getMaterial(col, repU, repV));
            obj.rotation.x = -Math.PI/2;    // in die Waagrechte drehen
            obj.scale.set(prXsiz, prYsiz, prZsiz);
            obj.position.set(x0+prXofs, y0+prYofs, z0+prZofs);
            if (protoName == "")
                grBase.add(obj);
            else
                grProto.add(obj);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 4) == "disc") {
            parms = splitParams(line.slice(5), 7);
            if (parms.length < 7) {
                addErrorMsg("disc: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            rad = rdFloatPar(parms[3]);
            r2 = rdFloatPar(parms[4]);
            nsid = rdIntPar(parms[5]);
            col = parms[6];
            geo = new THREE.RingGeometry(r2, rad, nsid);
            obj = new THREE.Mesh(geo, getMaterial(col));
            obj.rotation.x = -Math.PI/2;    // in die Waagrechte drehen
            obj.scale.set(prXsiz, prYsiz, prZsiz);
            obj.position.set(x0+prXofs, y0+prYofs, z0+prZofs);
            if (protoName == "")
                grBase.add(obj);
            else
                grProto.add(obj);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 8) == "boundbox") {
            parms = splitParams(line.slice(9), 6);
            if (parms.length < 6) {
                addErrorMsg("boundbox: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            wid = rdFloatPar(parms[3]);
            hei = rdFloatPar(parms[4]);
            len = rdFloatPar(parms[5]);
            if (snapMode == 1)
                y0 += hei/2*prYsiz;
            else if (snapMode == 2)
                y0 -= hei/2*prYsiz;
            else if (snapMode == 3)
                x0 += wid/2*prXsiz;
            else if (snapMode == 4)
                x0 -= wid/2*prXsiz;
            else if (snapMode == 5)
                z0 -= len/2*prZsiz;
            else if (snapMode == 6)
                z0 += len/2*prZsiz;
            addBBox(x0, y0-hei/2, z0, wid*prXsiz, hei*prYsiz, len*prZsiz);
            snapShift(wid, hei, len);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 4) == "cube") {
            parms = splitParams(line.slice(5), 9);
            if (parms.length < 7) {
                addErrorMsg("cube: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            wid = rdFloatPar(parms[3]);
            hei = rdFloatPar(parms[4]);
            len = rdFloatPar(parms[5]);
            col = parms[6];
            repU = 0;
            repV = 0;
            if (parms.length > 7) {
                repU = rdFloatPar(parms[7]);
                if (parms.length > 8)
                    repV = rdFloatPar(parms[8]);
                else
                    repV = repU;
            }
            geo = new THREE.BoxGeometry(wid, hei, len);
            addToGroup(geo, x0, y0, z0, wid/2, hei/2, len/2, col, repU, repV);
            if (protoName == "")
                addBBox(x0, y0-hei/2, z0, wid*prXsiz, hei*prYsiz, len*prZsiz);
            snapShift(wid, hei, len);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 4) == "wall") {
            parms = splitParams(line.slice(5), 9);
            if (parms.length < 7) {
                addErrorMsg("wall: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            wid = rdFloatPar(parms[3]);
            hei = rdFloatPar(parms[4]);
            len = rdFloatPar(parms[5]);
            col = parms[6];
            repU = 0;
            repV = 0;
            if (parms.length > 7) {
                repU = rdFloatPar(parms[7]);
                if (parms.length > 8)
                    repV = rdFloatPar(parms[8]);
                else
                    repV = repU;
            }
            geo = new THREE.BoxGeometry(wid, hei, len);
            addToGroup(geo, x0+wid/2, y0+hei/2, z0+len/2, 0, 0, 0, col, repU, repV);
            if (protoName == "")
                addBBox(x0+wid/2, y0, z0+len/2, wid*prXsiz, hei*prYsiz, len*prZsiz);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 7) == "sphere ") {
            parms = splitParams(line.slice(7), 8);
            if (parms.length < 5) {
                addErrorMsg("sphere: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            rad = rdFloatPar(parms[3]);
            repU = 0;
            repV = 0;
            if (parms.length < 6) {
                nsid = 12;
                col = parms[4];
            } else {
                nsid = rdIntPar(parms[4]);
                col = parms[5];
                if (parms.length > 6) {
                    repU = rdFloatPar(parms[6]);
                    if (parms.length > 7)
                        repV = rdFloatPar(parms[7]);
                    else
                        repV = repU;
                }
            }
            geo = new THREE.SphereGeometry(rad, nsid, nsid);
            addToGroup(geo, x0, y0, z0, rad, rad, rad, col, repU, repV);
            if (protoName == "")
                addBBox(x0, y0-rad, z0, rad*1.9*prXsiz, rad*1.9*prYsiz, rad*1.9*prZsiz);
            snapShift(rad*2, rad*2, rad*2);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 8) == "sphere1 ") {
            // sphere1 x0 y0 z0 rad wSeg hSeg phiBeg phiLen thetaBeg thetaLen col
            parms = splitParams(line.slice(8), 13);
            if (parms.length < 11) {
                addErrorMsg("sphere1: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            rad = rdFloatPar(parms[3]);
            var wSeg = rdIntPar(parms[4]);
            var hSeg = rdIntPar(parms[5]);
            var phiBeg = rdFloatPar(parms[6]) * Math.PI / 180;
            var phiLen = rdFloatPar(parms[7]) * Math.PI / 180;
            var theBeg = rdFloatPar(parms[8]) * Math.PI / 180;
            var theLen = rdFloatPar(parms[9]) * Math.PI / 180;
            col = parms[10];
            repU = 0;
            repV = 0;
            if (parms.length > 11) {
                repU = rdFloatPar(parms[11]);
                if (parms.length > 12)
                    repV = rdFloatPar(parms[12]);
                else
                    repV = repU;
            }
            geo = new THREE.SphereGeometry(rad, wSeg, hSeg, phiBeg, phiLen, theBeg, theLen);
            addToGroup(geo, x0, y0, z0, rad, rad, rad, col, repU, repV);
            if (protoName == "")
                addBBox(x0, y0-rad, z0, rad*1.9*prXsiz, rad*1.9*prYsiz, rad*1.9*prZsiz);
            snapShift(rad*2, rad*2, rad*2);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 8) == "cylinder") {
            parms = splitParams(line.slice(9), 9);
            if (parms.length < 6) {
                addErrorMsg("cylinder: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            hei = rdFloatPar(parms[3]);
            rad = rdFloatPar(parms[4]);
            repU = 0;
            repV = 0;
            if (parms.length < 7) {
                nsid = 12;
                col = parms[5];
            } else {
                nsid = rdIntPar(parms[5]);
                col = parms[6];
                if (parms.length > 7) {
                    repU = rdFloatPar(parms[7]);
                    if (parms.length > 8)
                        repV = rdFloatPar(parms[8]);
                    else
                        repV = repU;
                }
            }
            geo = new THREE.CylinderGeometry(rad, rad, hei, nsid, 2, bOpenEnded);
            addToGroup(geo, x0, y0, z0, rad, hei/2, rad, col, repU, repV);
            if (protoName == "")
                addBBox(x0, y0-hei/2, z0, rad*1.9*prXsiz, hei*prYsiz, rad*1.9*prZsiz);
            snapShift(rad*2, hei, rad*2);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 4) == "cone") {
            parms = splitParams(line.slice(5), 10);
            if (parms.length < 7) {
                addErrorMsg("cone: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            hei = rdFloatPar(parms[3]);
            r1 = rdFloatPar(parms[4]);
            r2 = rdFloatPar(parms[5]);
            repU = 0;
            repV = 0;
            if (parms.length < 8) {
                nsid = 12;
                col = parms[6];
            } else {
                nsid = rdIntPar(parms[6]);
                col = parms[7];
                if (parms.length > 8) {
                    repU = rdFloatPar(parms[8]);
                    if (parms.length > 9)
                        repV = rdFloatPar(parms[9]);
                    else
                        repV = repU;
                }
            }
            geo = new THREE.CylinderGeometry(r2, r1, hei, nsid, 2, bOpenEnded);
            addToGroup(geo, x0, y0, z0, r2, hei/2, r2, col, repU, repV);
            if (protoName == "") {
                var r0 = r1;
                if (r2 >= r1)
                    r0 = r2;
                addBBox(x0, y0-hei/2, z0, r0*1.9*prXsiz, hei*prYsiz, r0*1.9*prZsiz);
            }
            snapShift(r2, hei, r2);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "torus") {
            parms = splitParams(line.slice(6), 11);
            if (parms.length < 8) {
                addErrorMsg("torus: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            rad = rdFloatPar(parms[3]);
            tub = rdFloatPar(parms[4]);
            radSeg = rdIntPar(parms[5]);
            tubSeg = rdIntPar(parms[6]);
            repU = 0;
            repV = 0;
            if (parms.length < 9) {
                arc = Math.PI * 2;
                col = parms[7];
            } else {
                arc = rdFloatPar(parms[7]) * Math.PI / 180;
                col = parms[8];
                if (parms.length > 9) {
                    repU = rdFloatPar(parms[9]);
                    if (parms.length > 10)
                        repV = rdFloatPar(parms[10]);
                    else
                        repV = repU;
                }
            }
            if (snapMode == 1)
                y0 += tub*prYsiz;
            else if (snapMode == 2)
                y0 -= tub*prYsiz;
            else if (snapMode == 3)
                x0 += rad*prXsiz;
            else if (snapMode == 4)
                x0 -= rad*prXsiz;
            else if (snapMode == 5)
                z0 -= tub*prZsiz;
            else if (snapMode == 6)
                z0 += tub*prZsiz;
            geo = new THREE.TorusGeometry(rad, tub, radSeg, tubSeg, arc);
            obj = new THREE.Mesh(geo, getMaterial(col, repU, repV));
            obj.scale.set(prXsiz, prYsiz, prZsiz);
            obj.rotation.set(Math.PI/2+prXrot, prYrot, prZrot);
            obj.position.set(x0+prXofs, y0+prYofs, z0+prZofs);
            if (protoName == "")
                grBase.add(obj);
            else
               grProto.add(obj);
            if (protoName == "") {
                var r0 = r1;
                if (r2 >= r1)
                    r0 = r2;
                addBBox(x0, y0-tub, z0, rad*2*prXsiz, tub*2*prYsiz, rad*2*prZsiz);
            }
            snapShift(rad*2, tub*2, rad*2);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "lines") {
            parms = splitParams(line.slice(6), 35);
            var n = parms.length;
            if (n < 7) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Mindestens 2 Punkte und Farbkennung definieren");
                break;
            }
            if (((n - 1) % 6) != 0) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Anzahl Linien * 3 + 1 erwartet");
                break;
            }
            col = parseInt(parms[n-1], 16);
            for (var i = 0; i < n-1; i += 6) {
                geo = new THREE.Geometry();
                x0 = rdFloatPar(parms[i]);
                y0 = rdFloatPar(parms[i+1]);
                z0 = rdFloatPar(parms[i+2]);
                geo.vertices.push(new THREE.Vector3(x0+prXofs, y0+prYofs, z0+prZofs));
                x0 = rdFloatPar(parms[i+3]);
                y0 = rdFloatPar(parms[i+4]);
                z0 = rdFloatPar(parms[i+5]);
                geo.vertices.push(new THREE.Vector3(x0+prXofs, y0+prYofs, z0+prZofs));
                obj = new THREE.Line(geo, new THREE.LineBasicMaterial({color:col}));
                if (protoName == "")
                    grBase.add(obj);
                else
                    grProto.add(obj);
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "trifan") {
            parms = splitParams(line.slice(7), 35);
            var n = parms.length;
            if (n < 10) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Mindestens 3 Punkte und Farbkennung definieren");
                break;
            }
            if (((n - 1) % 3) != 0) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Anzahl Punkte * 3 + 1 erwartet");
                break;
            }
            col = parms[n-1];
            geo = new THREE.Geometry();
            var nPkt = (n-1) / 3;
            for (var i = 0; i < nPkt; i++) {
                x0 = rdFloatPar(parms[i*3]);
                y0 = rdFloatPar(parms[i*3+1]);
                z0 = rdFloatPar(parms[i*3+2]);
                geo.vertices.push(new THREE.Vector3(x0, y0, z0));
            }
            var nTri = nPkt - 2;
            for (var i = 1; i <= nTri; i++)
                geo.faces.push(new THREE.Face3(0, i, i+1));
            geo.computeFaceNormals();
            addToGroup(geo, 0, 0, 0, 0, 0, 0, col);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "prism") {
            parms = splitParams(line.slice(6), 35);
            var n = parms.length;
            if (n < 8) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Mindestens Hoehe, drei x/z-Punkte und Farbkennung definieren");
                break;
            }
            hei = parseFloat(parms[0]);
            col = parms[n-1];
            geo = new THREE.Geometry();
            var nPkt = (n - 2) / 2;
            var xpu = new Array();
            var zpu = new Array();
            var xmin = +1000000;
            var xmax = -1000000;
            var zmin = +1000000;
            var zmax = -1000000;
            for (var i = 0; i < nPkt; i++) {
                xpu.push(rdFloatPar(parms[i*2+1]));
                zpu.push(rdFloatPar(parms[i*2+2]));
                geo.vertices.push(new THREE.Vector3(xpu[i], hei, zpu[i]));
                if (xpu[i] < xmin)
                    xmin = xpu[i]; 
                if (xpu[i] > xmax)
                    xmax = xpu[i]; 
                if (zpu[i] < zmin)
                    zmin = zpu[i]; 
                if (zpu[i] > zmax)
                    zmax = zpu[i]; 
            }
            for (var i = 0; i < nPkt; i++) {
                geo.vertices.push(new THREE.Vector3(xpu[i], 0, zpu[i]));
            }
            var nTri = nPkt - 2;
            for (var i = 1; i <= nTri; i++) {
                geo.faces.push(new THREE.Face3(0, i, i+1));
                geo.faces.push(new THREE.Face3(nPkt, nPkt+i, nPkt+i+1));
            }
            for (var i = 0; i < nPkt; i++) {
                var i1 = (i+1) % nPkt;
                geo.faces.push(new THREE.Face3(i, nPkt+i, i1));
                geo.faces.push(new THREE.Face3(nPkt+i, i1, nPkt+i1));
            }
            geo.computeFaceNormals();
            addToGroup(geo, 0, 0, 0, 0, 0, 0, col);
            if (protoName == "") {
                wid = xmax-xmin;
                len = zmax-zmin;
                addBBox(xmin+wid/2, 0, zmin+len/2, wid, hei, len);
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 6) == "rotobj") {
            parms = splitParams(line.slice(7), 35);
            var n = parms.length;
            if (n < 7) {
                addErrorMsg("Anzahl Parameter " + n + " ungueltig: "
                    + "Mindestens <grad>, <resol>, zwei x/y-Punkte und Farbkennung definieren");
                break;
            }
            var grad = rdFloatPar(parms[0]) * Math.PI/180;
            var resol = rdIntPar(parms[1]);
            if (resol < 3) {
                addErrorMsg("Mindestens drei Winkelschritte <resol> erwartet");
                break;
            }
            col = parms[n-1];
            geo = new THREE.Geometry();
            var nPkt = (n - 3) / 2;
            var xpu = new Array();
            var ypu = new Array();
            var xmin = +1000000;
            var xmax = -1000000;
            var ymin = +1000000;
            var ymax = -1000000;
            for (var i = 0; i < nPkt; i++) {
                xpu.push(rdFloatPar(parms[i*2+2]));
                ypu.push(rdFloatPar(parms[i*2+3]));
                geo.vertices.push(new THREE.Vector3(xpu[i], ypu[i], 0));
                if (xpu[i] < xmin)
                    xmin = xpu[i]; 
                if (xpu[i] > xmax)
                    xmax = xpu[i]; 
                if (ypu[i] < ymin)
                    ymin = ypu[i]; 
                if (ypu[i] > ymax)
                    ymax = ypu[i]; 
            }
            var dwi = grad / resol;
            for (var i = 0; i < resol; i++) {
                var wi = i * dwi;
                dx = Math.sin(wi);
                dz = Math.cos(wi);
                for (var j = 0; j < nPkt; j++)
                    geo.vertices.push(new THREE.Vector3(xpu[j]*dx, ypu[j], xpu[j]*dz));
            }
            for (var i = 0; i < resol; i++) {
                var i1 = (i+1) % resol;
                for (var j = 0; j < nPkt-1; j++) {
                    var ni = i * nPkt;
                    var ni1 = i1 * nPkt;
                    geo.vertices.push(new THREE.Face3(ni+j, ni+j+1, ni1+j));
                    geo.vertices.push(new THREE.Face3(ni+j, ni+j+1, ni1+j+1));
                }
            }
            geo.computeFaceNormals();
            addToGroup(geo, 0, 0, 0, 0, 0, 0, col);
            if (protoName == "") {
                wid = xmax-xmin;
                len = zmax-zmin;
                addBBox(xmin+wid/2, 0, zmin+len/2, wid, hei, len);
            }
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 3) == "rod") {
            parms = splitParams(line.slice(4), 9);
            if (parms.length < 8) {
                addErrorMsg("rod: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            x1 = rdFloatPar(parms[3]);
            y1 = rdFloatPar(parms[4]);
            z1 = rdFloatPar(parms[5]);
            rad = rdFloatPar(parms[6]);
            dx = x1 - x0;
            dy = y1 - y0;
            dz = z1 - z0;
            len = Math.sqrt(dx*dx, dy*dy, dz*dz);
            if (parms.length > 8) {
                nsid = rdIntPar(parms[7]);
                col = parms[8];
            } else {
                nsid = 8;
                col = parms[7];
            }
            geo = new THREE.CylinderGeometry(rad, rad, len, nsid, 2, bOpenEnded);
            obj = new THREE.Mesh(geo, getMaterial(col, repU, repV));
            // Drehwinkel in der x/y-Ebene
            rx = getKursRad(dy, dz);
            ry = getKursRad(dx, dz);
            rz = getKursRad(dx, dy);
            obj.rotation.set(rx, ry, rz);
            obj.position.set(x0+dx/2, y0+dy/2, z0+dz/2);
            if (protoName == "")
                grBase.add(obj);
            else
                grProto.add(obj);
            continue;
        }

        //---------------------------------------------------------------------
        if (line.substr(0, 5) == "text ") {
            // text x0 y0 z0  xr yr zr  txtName [fontName] [styleName]
            parms = splitParams(line.slice(5), 9);
            if (parms.length < 7) {
                addErrorMsg("text: Zu wenige Parameter");
                break;
            }
            x0 = rdFloatPar(parms[0]);
            y0 = rdFloatPar(parms[1]);
            z0 = rdFloatPar(parms[2]);
            rx = rdFloatPar(parms[3]);
            ry = rdFloatPar(parms[4]);
            rz = rdFloatPar(parms[5]);
            var txt = findString(parms[6]);
            if (txt == "") {
                addErrorMsg("text: Name des Text-Strings nicht gefunden");
                break;
            }
            var font = "Bold 20px Arial";
            if (parms.length > 7) {
                font = findString(parms[7]);
                if (font == "") {
                    addErrorMsg("text: Name des Font-Strings nicht gefunden");
                    break;
                }
            }
            var style = "rgba(255,255,0,0.95)";
            if (parms.length > 8) {
                style = findString(parms[8]);
                if (style == "") {
                    addErrorMsg("text: Name des Style-Strings nicht gefunden");
                    break;
                }
            }
            //TODO
            obj = CanvasTextObject(txt, font, style);
//            obj.scale.set(prXsiz, prYsiz, prZsiz);
            obj.rotation.set(rx+prXrot, ry+prYrot, rz+prZrot);
            obj.position.set(x0+prXofs, y0+prYofs, z0+prZofs);
            obj.updateMatrix();
            if (protoName == "")
                grBase.add(obj);
            else
                grProto.add(obj);
            continue;
        }


//===============================================================================
// Experiment: Wegstrecken bzw. Gleissysteme verlegen

        if (line.substr(0, 5) == "track") {
            // track <typ> <nam> x0 y0 z0  xr yr zr
            if (bTrack) {
                addErrorMsg("track: Wegstrecke wurde bereits begonnen");
                break;
            }
            parms = splitParams(line.slice(6), 8);
            if (parms.length < 8) {
                addErrorMsg("track: Zu wenige Parameter");
                break;
            }
            var trk = new Array(0, parms[0], parms[1],
                rdFloatPar(parms[2]), rdFloatPar(parms[3]), rdFloatPar(parms[4]),
                rdFloatPar(parms[5]), rdFloatPar(parms[6]), rdFloatPar(parms[7]));
            vsTracks.push(trk);
            bTrack = true;
            continue;
        }

        if (line.substr(0, 8) == "endtrack") {
            if (!bTrack) {
                addErrorMsg("endtrack unerwartet: keine Wegstrecke begonnen");
                break;
            }
            var trk = new Array(1);
            vsTracks.push(trk);
            bTrack = false;
            continue;
        }

        if (line.substr(0, 8) == "straight") {
            if (!bTrack) {
                addErrorMsg("Keine Wegstrecke begonnen");
                break;
            }
            parms = splitParams(line.slice(9), 1);
            if (parms.length < 1) {
                addErrorMsg("straight: Parameter fehlt");
                break;
            }
            var trk = new Array(2, rdFloatPar(parms[0]));
            vsTracks.push(trk);
            continue;
        }

        if (line.substr(0, 5) == "curve") {
            if (!bTrack) {
                addErrorMsg("Keine Wegstrecke begonnen");
                break;
            }
            parms = splitParams(line.slice(6), 2);
            if (parms.length < 2) {
                addErrorMsg("straight: zu wenige Parameter");
                break;
            }
            var trk = new Array(3, rdFloatPar(parms[0]), rdFloatPar(parms[1]));
            vsTracks.push(trk);
            continue;
        }

        if (line.substr(0, 8) == "waypoint") {
            if (!bTrack) {
                addErrorMsg("Keine Wegstrecke begonnen");
                break;
            }
            parms = splitParams(line.slice(9), 1);
            if (parms.length < 1) {
                addErrorMsg("waypoint: Parameter fehlt");
                break;
            }
            var trk = new Array(4, rdFloatPar(parms[0]));
            vsTracks.push(trk);
            continue;
        }

        if (line.substr(0, 6) == "branch") {
            if (!bTrack) {
                addErrorMsg("Keine Wegstrecke begonnen");
                break;
            }
            parms = splitParams(line.slice(7), 2);
            if (parms.length < 2) {
                addErrorMsg("branch: zu wenige Parameter");
                break;
            }
            var trk = new Array(5, rdFloatPar(parms[0]), rdFloatPar(parms[1]));
            vsTracks.push(trk);
            continue;
        }

        if (line.substr(0, 7) == "connect") {
            if (!bTrack) {
                addErrorMsg("Keine Wegstrecke begonnen");
                break;
            }
            parms = splitParams(line.slice(8), 1);
            if (parms.length < 1) {
                addErrorMsg("connect: Parameter fehlt");
                break;
            }
            var trk = new Array(6, rdFloatPar(parms[0]));
            vsTracks.push(trk);
            continue;
        }
//===============================================================================


        addErrorMsg("Unbekannte VSC-Anweisung: " + line);
        break;
    }

    if ((errMsg == "") && (protoName != ""))
        addErrorMsg("Script-Ende unerwartet: kein endproto fuer " + protoName);
    if (errMsg != "")
        alert(errMsg);

    // Die Datenlisten werden in einem geschlossenen Array zusammengefasst und uebergeben.

    var vscData = new Array(10);
    vscData[0] = vsName;                    // Dateiname
    vscData[1] = vsVersion;
    vscData[2] = grBase;
    vscData[3] = vsBoxes;
    vscData[4] = showBoundBoxes(vsBoxes);   // Wireframes
    vscData[5] = vsSymbols;
    vscData[6] = vsDynamics;
    vscData[7] = vsSensors;
    vscData[8] = vsCameras;
    vscData[9] = vsTracks;

    return vscData;
  }     // Ende doBuildVscScene()
 };     // Ende return{}
};      // Ende VscBuilder()


//=============================================================================
// Externer Aufruf kompatibel zur bisherigen Methode.

var buildVscScene = function(vsName, vscText) {
    var vscBuilder = new VscBuilder();
    var vscData = vscBuilder.doBuildVscScene(vsName, vscText);
    return vscData;
};

