//-----------------------------------------------------------------------------

var SceneVersion = "w2.js V16.02.06";
var SceneName = "Random-Labyrinth";

var SceneObj  = undefined;
var SceneId;

var xDim = 24;            // Anzahl Kacheln in x-Richtung
var zDim = 24;            // Anzahl Kacheln in z-Richtung
var aWid = 5;            // Ausdehnung einer Kachel
var wCube = aWid-0.4;    // Abmessungen eines Quaders
var lCube = aWid-0.4;    // Luecken zum Durchschauen lassen
var hCube = 10;

var sArray;

var oPosX = 0;          // Position eines beweglichen Objekts
var oPosY = 0;
var oPosZ = 0;
var oVx = 0;            // Geschwindigkeitskomponenten
var oVz = 0;

//-----------------------------------------------------------------------------

var buildScene = function(sid) {

    console.log("buildScene(" + sid + ")");

    SceneId = sid;
    SceneObj = new THREE.Mesh();

    SceneObj.add(buildGround('../tex/grasslight.jpg'));

    sArray = new Array(xDim);
    for (var i = 0; i < xDim; i++)
        sArray[i] = new Array(zDim);

    for (var i = 0; i < xDim; i++) {
        for (var j = 0; j < zDim; j++) {
            // freie Flaeche in der Mitte
            if ((i > xDim/2-3) && (i < xDim/2+3) && (j > zDim/2-3) && (j < zDim/2+3))
                sArray[i][j] = 0;
            else if (Math.random() > 0.3)
                sArray[i][j] = 0;
            else
                sArray[i][j] = 1 + Math.floor(Math.random() * 4);
            // An den Raendern die Mauern dicht machen
            if ((i == 0) || /*(i == xDim-1) ||*/ (j == 0) || (j == zDim-1))
                sArray[i][j] = 1;
        }
    }

    while (1) {
        var ox = Math.floor(Math.random() * xDim);         
        var oz = Math.floor(Math.random() * zDim);         
        if ((ox > xDim/2-3) && (ox < xDim/2+3) && (oz > zDim/2-3) && (oz < zDim/2+3))
            continue;                   // nicht in der Mitte
        if (sArray[ox][oz] == 0)
            break;                      // ok, hier freier Platz
    }
    oPosX = ox * aWid + aWid/2; 
    oPosZ = oz * aWid + aWid/2;
    oPosY = aWid/3;
    oVx = 1;
    oVz = 0;

    var cubeMap = THREE.ImageUtils.loadTexture('../tex/fliesen2.jpg');
    cubeMap.wrapS = cubeMap.wrapT = THREE.RepeatWrapping;
    cubeMap.anisotropy = 16;

    var cubeGeo = new THREE.CubeGeometry(wCube, hCube, lCube);
    cubeGeo.applyMatrix(new THREE.Matrix4().makeTranslation(0, hCube*0.46, 0));

    var cubeMat = new THREE.MeshLambertMaterial({map: cubeMap, side: THREE.DoubleSide});
    var aryMesh = new THREE.Mesh(cubeGeo, cubeMat);

    var aryGeo = new THREE.Geometry();

    for (var i = 0; i < xDim; i++) {
        for (var j = 0; j < zDim; j++) {
            var v = sArray[i][j];
            if (v != 0) {
                aryMesh.position.x = aWid*(i-xDim/2) + aWid/2;
                aryMesh.position.z = aWid*(j-zDim/2) + aWid/2;
                aryMesh.rotation.y = (v-1) * Math.PI/2;    // in 90-Grad-Stufen
                aryMesh.updateMatrix();
                aryGeo.merge(aryMesh.geometry, aryMesh.matrix);
            }
        }
    }
    SceneObj.add(new THREE.Mesh(aryGeo, cubeMat));

    SceneObj.add(new xWald());      // in w21.js

    return SceneObj;
};

//-----------------------------------------------------------------------------

function recvSceneData(rxText) {
}

//-----------------------------------------------------------------------------

function terminateScene() {
    console.log("terminateScene()");
    if (SceneObj !== undefined)
        delete SceneObj;
}

//-----------------------------------------------------------------------------

function cycleScene(msPerFrame) {

    var nxtPosX = oPosX + oVx*0.001;
    var nxtPosZ = oPosZ + oVz*0.001;
    if (!testScenePosition(nxtPosX-1.8, oPosY, nxtPosZ-1.8)) {
        if ((oVx > 0) && (oVz == 0)) {  // nach rechts unterwegs?
            oVx = 0;                    // nach unten probieren
            oVz = +1;
        } else if ((oVx == 0) && (oVz > 0)) { // nach unten?
            oVx = -1;                   // jetzt nach links probieren
            oVz = 0;
        } else if ((oVx == 0) && (oVz > 0)) { // nach links?
            oVx = 0;                    // jetzt nach oben probieren
            oVz = -1;
        } else {                        // nach oben unterwegs
            oVx = 0;                    // jetzt nach rechts probieren
            oVz = -1;
        }
    }
}

//-----------------------------------------------------------------------------

function testScenePosition (xp, yp, zp) {

    if ((yp < 0) || (yp > hCube))       // ober- oder unterhalb ?
        return true;                    // frei

    var x = Math.floor(xp/aWid) + xDim/2;
    var z = Math.floor(zp/aWid) + zDim/2;
    if ((x < 0) || (x >= xDim) || (z < 0) || (z >= zDim))
        return true;                    // frei ausserhalb des Arrays

    if (sArray[x][z] == 0)
        return true;                    // freie Position
    else
        return false;
}

//-----------------------------------------------------------------------------

function getSceneGroundLevel(xp, yp, zp) {
    return 0;
}
